from resolwe.process import IntegerField, Process, StringField


class PythonProcessDataIdBySlug(Process):
    """The process is used for testing get_data_id_by_slug."""

    slug = "test-python-process-data-id-by-slug"
    name = "Test Python Process Data ID by Slug"
    version = "1.0.0"
    process_type = "data:python:dataidbyslug"
    requirements = {
        "executor": {
            "docker": {
                "image": "resolwe/base:ubuntu-18.04",
            }
        },
        "resources": {
            "network": True,
        },
    }

    class Input:
        """Input fields."""

        slug = StringField(label="Slug")

    class Output:
        data_id = IntegerField(label="Data ID")

    def run(self, inputs, outputs):
        data_id = self.get_data_id_by_slug(inputs.slug)
        outputs.data_id = data_id



"""Resolwe migration history models."""
from django.contrib.postgres.fields import JSONField
from django.db import models


class MigrationHistoryBase(models.Model):
    """Abstract model for storing migration history."""

    class Meta:
        """Model meta."""

        abstract = True

    #: migration identifier
    migration = models.CharField(max_length=255, db_index=True)

    #: creation date and time
    created = models.DateTimeField(auto_now_add=True)

    #: migration-specific metadata
    metadata = JSONField(default=dict)

    def __str__(self):
        """Format model name."""
        return self.migration


class ProcessMigrationHistory(MigrationHistoryBase):
    """Model for storing process migration history."""

    process = models.ForeignKey(
        "Process", related_name="migration_history", on_delete=models.CASCADE
    )


class DataMigrationHistory(MigrationHistoryBase):
    """Model for storing data migration history."""

    data = models.ForeignKey(
        "Data", related_name="migration_history", on_delete=models.CASCADE
    )



""".. Ignore pydocstyle D400.

=====
State
=====

Synchronized singleton state container for the manager.

.. autofunction:: resolwe.flow.managers.state.update_constants
.. autoclass:: resolwe.flow.managers.state.ManagerState

"""

# This module should not import anything local, or there will be circular
# dependencies, since the constants are needed in various sub-modules inside
# resolwe.flow.managers.

import json
import os
from collections import namedtuple
from pathlib import PurePath

import redis

from django.conf import settings

ManagerChannelPair = namedtuple("ManagerChannelPair", ["queue", "queue_response"])

MANAGER_CONTROL_CHANNEL = "DUMMY.control"
MANAGER_EXECUTOR_CHANNELS = ManagerChannelPair("DUMMY.queue", "DUMMY.queue_response")
MANAGER_STATE_PREFIX = "DUMMY.state_prefix"
MANAGER_LISTENER_STATS = "DUMMY.listener_stats"


def update_constants():
    """Recreate channel name constants with changed settings.

    This kludge is mostly needed due to the way Django settings are
    patched for testing and how modules need to be imported throughout
    the project. On import time, settings are not patched yet, but some
    of the code needs static values immediately. Updating functions such
    as this one are then needed to fix dummy values.
    """
    global MANAGER_CONTROL_CHANNEL, MANAGER_EXECUTOR_CHANNELS
    global MANAGER_LISTENER_STATS, MANAGER_STATE_PREFIX
    redis_prefix = getattr(settings, "FLOW_MANAGER", {}).get("REDIS_PREFIX", "")

    MANAGER_CONTROL_CHANNEL = "{}.control".format(redis_prefix)
    MANAGER_EXECUTOR_CHANNELS = ManagerChannelPair(
        "{}.result_queue".format(redis_prefix),
        "{}.result_queue_response".format(redis_prefix),
    )
    MANAGER_STATE_PREFIX = "{}.state".format(redis_prefix)
    MANAGER_LISTENER_STATS = "{}.listener_stats".format(redis_prefix)


update_constants()


class PathEncoder(json.JSONEncoder):
    """Encode objects of type PurePath as strings."""

    def default(self, obj):
        """Override default method."""
        return os.fspath(obj) if isinstance(obj, PurePath) else super().default(obj)


class ManagerState:
    """State interface implementation.

    This holds variables required to be shared between all manager
    workers and takes care of operation atomiticy and synchronization.
    Redis facilitates storage shared between workers, whereas
    atomicity needs to be dealt with explicitly; this interface hides
    the Redis and Python details required to achieve
    syntax-transparent atomicity (such as being able to do
    ``executor_count += 1``, a load-modify-store operation sequence).
    """

    class LuaFunction:
        """Wrapper class for Lua function strings."""

        def __init__(self, owner, script, short=False):
            """Construct a Lua function instance and register it.

            :param owner: The owning :class:`RedisAtomicBase` instance.
            :param script: The Lua script text to register.
            :param short: If ``True``, the script as given is just an
                expression that still needs to be wrapped into a full
                script.
            """
            if short:
                self.script = (
                    """
                    local oldval = tonumber(redis.call('EXISTS', KEYS[1]) and redis.call('GET', KEYS[1]) or ARGV[2])
                    local arg = tonumber(ARGV[1])
                    local newval = """
                    + script
                    + """
                    redis.call('SET', KEYS[1], newval)
                    return newval
                """
                )
            else:
                self.script = script

            self.owner = owner
            self.function = owner.redis.register_script(self.script)

        def __call__(self, *args):
            """Call the underlying Lua function transparently."""
            return self.function(keys=[self.owner.item_name], args=args)

    class RedisAtomicBase:
        """Base class for atomic data handling."""

        def __init__(self, conn, key_prefix, item_name, *args, **kwargs):
            """Initialize the base instance, register Lua scripts.

            :param conn: The Redis interface object.
            :param key_prefix: The key prefix used for variable names.
            :param item_name: The name of this variable.
            """
            self.redis = conn
            self.item_name = key_prefix + "." + item_name

    class IntegerDatum(RedisAtomicBase):
        """Class for integer data supporting standard arithmetic."""

        def __init__(self, *args, **kwargs):
            """Initialize atomic integer instance.

            :param initial_value: Optional. The initial value of this
                variable in case it doesn't exist in Redis yet.
            """
            super().__init__(*args, **kwargs)
            self._lua_add = ManagerState.LuaFunction(self, "oldval + arg", short=True)
            self._lua_mul = ManagerState.LuaFunction(self, "oldval * arg", short=True)
            self._lua_floordiv = ManagerState.LuaFunction(
                self, "floor(oldval / arg)", short=True
            )
            self._lua_cas = ManagerState.LuaFunction(
                self,
                """
                local oldval = tonumber(redis.call('EXISTS', KEYS[1]) and redis.call('GET', KEYS[1]) or ARGV[3])
                if oldval == tonumber(ARGV[1]) then
                    redis.call('SET', KEYS[1], tonumber(ARGV[2]))
                    return 1
                else
                    return 0
                end
            """,
            )
            self.initial_value = kwargs.get("initial_value", 0)

        def __int__(self):
            """Convert the proxy into an integer."""
            # Use incrby instead of get, as incrby always returns an integer.
            return self.redis.incrby(self.item_name, 0)

        def set(self, val=None):
            """Set to a new value."""
            self.redis.set(self.item_name, val if val else self.initial_value)

        def add(self, val):
            """Add a value to the integer atomically."""
            return self._lua_add(val, self.initial_value)

        def mul(self, val):
            """Multiply atomically by the given value."""
            return self._lua_mul(val, self.initial_value)

        def floordiv(self, val):
            """Perform integer division atomically."""
            return self._lua_floordiv(val, self.initial_value)

        def cas(self, oldval, newval):
            """Compare and swap operation."""
            return self._lua_cas(oldval, newval, self.initial_value)

    class ObjectDatum(RedisAtomicBase):
        """Redis atomic class for JSONifiable Python data."""

        def set(self, val):
            """Set to a new value.

            :param val: The value to be serialized into JSON.
            """
            val = json.dumps(val, cls=PathEncoder)
            self.redis.set(self.item_name, val)

        def get(self):
            """Parse the value from the internal JSON representation.

            :return: The deserialized python object represented by the
                JSON in this variable, or ``None``.
            """
            val = self.redis.get(self.item_name)
            if not val:
                return None
            try:
                val = json.loads(val.decode("utf-8"))
            except json.JSONDecodeError:
                val = None
            return val

    def __init__(self, key_prefix):
        """Initialize the Redis connection.

        :param key_prefix: The key prefix used for variable names.
        """
        self.redis = redis.StrictRedis(
            **getattr(settings, "FLOW_EXECUTOR", {}).get("REDIS_CONNECTION", {})
        )
        self.key_prefix = key_prefix
        self._settings_override = self.ObjectDatum(
            self.redis, key_prefix, "settings_override"
        )

    def reset(self):
        """Reset all properties to their initial values."""
        self.settings_override = None

    def destroy_channels(self):
        """Destroy Redis channels managed by this state instance."""
        for item_name in dir(self):
            item = getattr(self, item_name)
            if isinstance(item, self.RedisAtomicBase):
                self.redis.delete(item.item_name)

    @property
    def settings_override(self):
        """Get the settings overrides for the manager workers."""
        return self._settings_override.get()

    @settings_override.setter
    def settings_override(self, newval):
        """Set a new settings override object."""
        self._settings_override.set(newval)



"""An execution engine for Python processes."""
import copy
import inspect
import json
import os
import shlex
import shutil

from resolwe.flow.execution_engines.base import BaseExecutionEngine
from resolwe.flow.models.utils import (
    get_collection_of_input_entities,
    hydrate_input_references,
    hydrate_input_uploads,
    serialize_collection_relations,
)
from resolwe.process.parser import SafeParser

PYTHON_RUNTIME_DIRNAME = "python_runtime"
PYTHON_RUNTIME_ROOT = "/"
PYTHON_RUNTIME_VOLUME = os.path.join(PYTHON_RUNTIME_ROOT, PYTHON_RUNTIME_DIRNAME)
PYTHON_PROGRAM_ROOT = "/"
PYTHON_PROGRAM_FILENAME = "python_process.py"
PYTHON_PROGRAM_VOLUME = os.path.join(PYTHON_PROGRAM_ROOT, PYTHON_PROGRAM_FILENAME)
PYTHON_INPUTS_FILENAME = "inputs.json"
PYTHON_INPUTS_ROOT = "/"
PYTHON_INPUTS_VOLUME = os.path.join(PYTHON_INPUTS_ROOT, PYTHON_INPUTS_FILENAME)
PYTHON_RELATIONS_FILENAME = "relations.json"
PYTHON_RELATIONS_ROOT = "/"
PYTHON_RELATIONS_VOLUME = os.path.join(PYTHON_RELATIONS_ROOT, PYTHON_RELATIONS_FILENAME)
PYTHON_REQUIREMENTS_FILENAME = "requirements.json"
PYTHON_REQUIREMENTS_ROOT = "/"
PYTHON_REQUIREMENTS_VOLUME = os.path.join(
    PYTHON_INPUTS_ROOT, PYTHON_REQUIREMENTS_FILENAME
)


class ExecutionEngine(BaseExecutionEngine):
    """An execution engine that outputs bash programs."""

    name = "python"

    def discover_process(self, path):
        """Perform process discovery in given path.

        This method will be called during process registration and
        should return a list of dictionaries with discovered process
        schemas.
        """
        if not path.lower().endswith(".py"):
            return []

        parser = SafeParser(open(path).read())
        processes = parser.parse()
        return [process.to_schema() for process in processes]

    def evaluate(self, data):
        """Evaluate the code needed to compute a given Data object."""
        return (
            'PYTHONPATH="{runtime}" python3 -u -m resolwe.process {program} '
            "--slug {slug} "
            "--name {name} "
            "--inputs {inputs} "
            "--relations {relations} "
            "--requirements {requirements}".format(
                runtime=PYTHON_RUNTIME_VOLUME,
                program=PYTHON_PROGRAM_VOLUME,
                slug=shlex.quote(data.process.slug),
                name=shlex.quote(data.name),
                inputs=PYTHON_INPUTS_VOLUME,
                relations=PYTHON_RELATIONS_VOLUME,
                requirements=PYTHON_REQUIREMENTS_VOLUME,
            )
        )

    def prepare_runtime(self, runtime_dir, data):
        """Prepare runtime directory."""
        # Copy over Python process runtime (resolwe.process).
        import resolwe.process as runtime_package

        src_dir = os.path.dirname(inspect.getsourcefile(runtime_package))
        dest_package_dir = os.path.join(
            runtime_dir, PYTHON_RUNTIME_DIRNAME, "resolwe", "process"
        )
        shutil.copytree(src_dir, dest_package_dir)
        os.chmod(dest_package_dir, 0o755)

        # Write python source file.
        source = data.process.run.get("program", "")
        program_path = os.path.join(runtime_dir, PYTHON_PROGRAM_FILENAME)
        with open(program_path, "w") as file:
            file.write(source)
        os.chmod(program_path, 0o755)

        # Write serialized inputs.
        inputs = copy.deepcopy(data.input)
        hydrate_input_references(inputs, data.process.input_schema)
        hydrate_input_uploads(inputs, data.process.input_schema)
        inputs_path = os.path.join(runtime_dir, PYTHON_INPUTS_FILENAME)

        # XXX: Skip serialization of LazyStorageJSON. We should support
        # LazyStorageJSON in Python processes on the new communication protocol
        def default(obj):
            """Get default value."""
            class_name = obj.__class__.__name__
            if class_name == "LazyStorageJSON":
                return ""

            raise TypeError(f"Object of type {class_name} is not JSON serializable")

        with open(inputs_path, "w") as file:
            json.dump(inputs, file, default=default)

        # Write serialized requirements.
        # Include special 'requirements' variable in the context.
        requirements = copy.deepcopy(data.process.requirements)
        # Inject default values and change resources according to
        # the current Django configuration.
        requirements["resources"] = data.process.get_resource_limits()
        requirements_path = os.path.join(runtime_dir, PYTHON_REQUIREMENTS_FILENAME)

        with open(requirements_path, "w") as file:
            json.dump(requirements, file)

        # Write serialized relations
        relations = {}
        if "relations" in requirements:
            collection = get_collection_of_input_entities(data)
            relations = serialize_collection_relations(collection)
        relations_path = os.path.join(runtime_dir, PYTHON_RELATIONS_FILENAME)

        with open(relations_path, "w") as file:
            json.dump(relations, file)

        # Generate volume maps required to expose needed files.
        volume_maps = {
            PYTHON_RUNTIME_DIRNAME: PYTHON_RUNTIME_VOLUME,
            PYTHON_PROGRAM_FILENAME: PYTHON_PROGRAM_VOLUME,
            PYTHON_INPUTS_FILENAME: PYTHON_INPUTS_VOLUME,
            PYTHON_RELATIONS_FILENAME: PYTHON_RELATIONS_VOLUME,
            PYTHON_REQUIREMENTS_FILENAME: PYTHON_REQUIREMENTS_VOLUME,
        }

        return volume_maps



"""Connector utils."""
import concurrent.futures
import os
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

from .hasher import compute_hashes

if TYPE_CHECKING:
    from os import PathLike

T = TypeVar("T")


def get_transfer_object(
    file_: "PathLike[str]", base_path: "PathLike[str]"
) -> Optional[Dict[str, Union[str, int]]]:
    """Get transfer object from file/directory.

    :param file_: path-like object pointing to a file/directory.

    :param base_path: path in the returned dictionary will be relative with
        respect to this path.

    :returns: dictionary that can be used as an input for transfer methods.
        If file_ argument does not point to a file/directory None is
        returned.
    """
    path = Path(file_)
    if not (path.is_file() or path.is_dir()):
        return None

    file_transfer_data = {
        "size": path.stat().st_size,
        "path": os.fspath(path.relative_to(base_path)),
    }
    # Append '/' to directory path.
    if path.is_dir():
        file_transfer_data["path"] = os.path.join(file_transfer_data["path"], "")
        file_transfer_data["size"] = 0

    file_transfer_data.update(compute_hashes(path))
    return file_transfer_data


def chunks(iterable: Sequence[T], n: int) -> Iterable[Sequence[T]]:
    """Split iterable into n evenly sized chunks."""
    for i in range(0, n):
        yield iterable[i::n]


def paralelize(
    objects: Sequence[Any],
    worker: Callable[[Sequence[Any]], Any],
    max_threads: int = 10,
) -> Sequence[concurrent.futures.Future]:
    """Paralelize tasks using connector on list of URLS.

    URLs are split into up-to num_threads chunks and each chunk is processed
    in its own thread. Connectors in worker method MUST be duplicated to ensure
    thread safety.

    :returns: collection of instance of Future objects, each one corresponding
        to one thread. It is caller responsibility to check if threads have
        finished successfully.
    """
    number_of_chunks = min(len(objects), max_threads)
    objects_chunks = chunks(objects, number_of_chunks)

    futures = []
    with concurrent.futures.ThreadPoolExecutor(max_workers=max_threads) as executor:
        for objects_chunk in objects_chunks:
            futures.append(executor.submit(worker, objects_chunk))
    return futures



""".. Ignore pydocstyle D400.

===============
Local Connector
===============

"""
import logging
import subprocess

from resolwe.utils import BraceMessage as __

from .base import BaseConnector

logger = logging.getLogger(__name__)


class Connector(BaseConnector):
    """Local connector for job execution."""

    def submit(self, data, runtime_dir, argv):
        """Run process locally.

        For details, see
        :meth:`~resolwe.flow.managers.workload_connectors.base.BaseConnector.submit`.
        """
        logger.debug(
            __(
                "Connector '{}' running for Data with id {} ({}).",
                self.__class__.__module__,
                data.id,
                repr(argv),
            )
        )
        subprocess.Popen(argv, cwd=runtime_dir, stdin=subprocess.DEVNULL).wait()



# pylint: disable=missing-docstring
from pathlib import Path
from unittest.mock import MagicMock, call, patch

from resolwe.flow.managers.listener import ExecutorListener
from resolwe.flow.managers.protocol import ExecutorProtocol
from resolwe.flow.models import Data, DataDependency
from resolwe.storage.models import FileStorage, ReferencedPath, StorageLocation
from resolwe.test import TestCase


class ListenerTest(TestCase):
    fixtures = ["storage_data.yaml", "storage_processes.yaml", "storage_users.yaml"]

    @classmethod
    def setUpTestData(cls):
        cls.listener = ExecutorListener()
        cls.file_storage = FileStorage.objects.get(id=1)
        cls.storage_location = StorageLocation.objects.create(
            file_storage=cls.file_storage, connector_name="GCS", status="OK"
        )
        cls.path = ReferencedPath.objects.create(
            path="test.me", md5="md5", crc32c="crc", awss3etag="aws"
        )
        cls.storage_location.files.add(cls.path)

    @patch("resolwe.flow.managers.listener.ExecutorListener._send_reply")
    @patch("resolwe.flow.managers.listener.async_to_sync")
    def test_handle_download_finished_missing_storage_location(
        self, async_to_sync_mock, send_reply_mock
    ):
        obj = {
            "command": ExecutorProtocol.DOWNLOAD_FINISHED,
            "data_id": -1,
            "storage_location_id": -2,
        }
        send_wrapper = MagicMock()
        async_to_sync_mock.return_value = send_wrapper
        with patch(
            "resolwe.storage.models.FileStorage.default_storage_location",
            self.storage_location,
        ):
            self.listener.handle_download_finished(obj)

        async_to_sync_mock.assert_called_once_with(send_reply_mock)
        send_wrapper.assert_called_once_with(
            {
                "command": "download_finished",
                "data_id": -1,
                "storage_location_id": -2,
            },
            {"result": "ER"},
        )

    @patch("resolwe.flow.managers.listener.ExecutorListener._send_reply")
    @patch("resolwe.flow.managers.listener.async_to_sync")
    def test_handle_download_finished(self, async_to_sync_mock, send_reply_mock):
        storage_location = StorageLocation.objects.create(
            file_storage=self.file_storage, connector_name="local"
        )
        obj = {
            "command": ExecutorProtocol.DOWNLOAD_FINISHED,
            "data_id": 1,
            "storage_location_id": storage_location.id,
        }
        send_wrapper = MagicMock()
        async_to_sync_mock.return_value = send_wrapper

        with patch(
            "resolwe.storage.models.FileStorage.default_storage_location",
            self.storage_location,
        ):
            self.listener.handle_download_finished(obj)

        send_wrapper.assert_called_once_with(
            {
                "command": "download_finished",
                "data_id": 1,
                "storage_location_id": storage_location.id,
            },
            {"result": "OK"},
        )
        storage_location.refresh_from_db()
        self.assertEqual(storage_location.status, StorageLocation.STATUS_DONE)
        self.assertEqual(storage_location.files.count(), 1)
        file = storage_location.files.get()
        self.assertEqual(file.path, "test.me")
        self.assertEqual(file.md5, "md5")
        self.assertEqual(file.crc32c, "crc")
        self.assertEqual(file.awss3etag, "aws")

    @patch("resolwe.flow.managers.listener.logger.error")
    def test_handle_download_aborted_missing_storage_location(self, error_logger_mock):
        obj = {
            "command": ExecutorProtocol.DOWNLOAD_ABORTED,
            "data_id": -1,
            "storage_location_id": -2,
        }

        self.listener.handle_download_aborted(obj)
        error_logger_mock.assert_called_once_with(
            "StorageLocation for data does not exist",
            extra={"storage_location_id": -2, "data_id": -1},
        )

    @patch("resolwe.flow.managers.listener.logger.error")
    def test_handle_download_aborted(self, error_logger_mock):
        storage_location = StorageLocation.objects.create(
            file_storage=self.file_storage,
            connector_name="local",
            status=StorageLocation.STATUS_UPLOADING,
        )
        obj = {
            "command": ExecutorProtocol.DOWNLOAD_ABORTED,
            "data_id": -1,
            "storage_location_id": storage_location.id,
        }
        self.listener.handle_download_aborted(obj)
        error_logger_mock.assert_not_called()
        storage_location.refresh_from_db()
        self.assertEqual(storage_location.status, StorageLocation.STATUS_PREPARING)

    @patch("resolwe.flow.managers.listener.consumer.send_event")
    @patch("resolwe.flow.managers.listener.logger.error")
    @patch("resolwe.flow.managers.listener.ExecutorListener._send_reply")
    @patch("resolwe.flow.managers.listener.async_to_sync")
    def test_handle_download_started_no_location(
        self, async_to_sync_mock, send_reply_mock, error_logger_mock, send_event_mock
    ):
        obj = {
            "command": ExecutorProtocol.DOWNLOAD_STARTED,
            "data_id": -1,
            "storage_location_id": -2,
            "download_started_lock": True,
        }
        self.listener.handle_download_started(obj)

        async_to_sync_mock.assert_has_calls(
            [
                call(send_reply_mock),
                call()(
                    {
                        "command": "download_started",
                        "data_id": -1,
                        "storage_location_id": -2,
                        "download_started_lock": True,
                    },
                    {"result": "ER"},
                ),
                call(send_event_mock),
                call()(
                    {
                        "command": "abort_data",
                        "data_id": -1,
                        "communicate_kwargs": {
                            "executor": "resolwe.flow.executors.local"
                        },
                    }
                ),
            ]
        )
        error_logger_mock.assert_called_once_with(
            "StorageLocation for downloaded data does not exist",
            extra={"storage_location_id": -2},
        )

    @patch("resolwe.flow.managers.listener.consumer.send_event")
    @patch("resolwe.flow.managers.listener.logger.error")
    @patch("resolwe.flow.managers.listener.ExecutorListener._send_reply")
    @patch("resolwe.flow.managers.listener.async_to_sync")
    def test_handle_download_started_ok_no_lock_preparing(
        self, async_to_sync_mock, send_reply_mock, error_logger_mock, send_event_mock
    ):
        storage_location = StorageLocation.objects.create(
            file_storage=self.file_storage, connector_name="local"
        )

        obj = {
            "command": ExecutorProtocol.DOWNLOAD_STARTED,
            "data_id": -1,
            "storage_location_id": storage_location.id,
            "download_started_lock": False,
        }
        self.listener.handle_download_started(obj)

        async_to_sync_mock.assert_has_calls(
            [
                call(send_reply_mock),
                call()(
                    {
                        "command": "download_started",
                        "data_id": -1,
                        "storage_location_id": storage_location.id,
                        "download_started_lock": False,
                    },
                    {"result": "OK", "download_result": "download_started"},
                ),
            ]
        )
        error_logger_mock.assert_not_called()
        storage_location.refresh_from_db()
        self.assertEqual(storage_location.status, StorageLocation.STATUS_PREPARING)

    @patch("resolwe.flow.managers.listener.consumer.send_event")
    @patch("resolwe.flow.managers.listener.logger.error")
    @patch("resolwe.flow.managers.listener.ExecutorListener._send_reply")
    @patch("resolwe.flow.managers.listener.async_to_sync")
    def test_handle_download_started_ok_no_lock_uploading(
        self, async_to_sync_mock, send_reply_mock, error_logger_mock, send_event_mock
    ):
        storage_location = StorageLocation.objects.create(
            file_storage=self.file_storage,
            connector_name="local",
            status=StorageLocation.STATUS_UPLOADING,
        )
        obj = {
            "command": ExecutorProtocol.DOWNLOAD_STARTED,
            "data_id": -1,
            "storage_location_id": storage_location.id,
            "download_started_lock": False,
        }

        self.listener.handle_download_started(obj)

        async_to_sync_mock.assert_has_calls(
            [
                call(send_reply_mock),
                call()(
                    {
                        "command": "download_started",
                        "data_id": -1,
                        "storage_location_id": storage_location.id,
                        "download_started_lock": False,
                    },
                    {"result": "OK", "download_result": "download_in_progress"},
                ),
            ]
        )
        error_logger_mock.assert_not_called()
        storage_location.refresh_from_db()
        self.assertEqual(storage_location.status, StorageLocation.STATUS_UPLOADING)

    @patch("resolwe.flow.managers.listener.consumer.send_event")
    @patch("resolwe.flow.managers.listener.logger.error")
    @patch("resolwe.flow.managers.listener.ExecutorListener._send_reply")
    @patch("resolwe.flow.managers.listener.async_to_sync")
    def test_handle_download_started_ok_no_lock_done(
        self, async_to_sync_mock, send_reply_mock, error_logger_mock, send_event_mock
    ):
        storage_location = StorageLocation.objects.create(
            file_storage=self.file_storage,
            connector_name="local",
            status=StorageLocation.STATUS_DONE,
        )
        obj = {
            "command": ExecutorProtocol.DOWNLOAD_STARTED,
            "data_id": -1,
            "storage_location_id": storage_location.id,
            "download_started_lock": False,
        }
        self.listener.handle_download_started(obj)
        async_to_sync_mock.assert_has_calls(
            [
                call(send_reply_mock),
                call()(
                    {
                        "command": "download_started",
                        "data_id": -1,
                        "storage_location_id": storage_location.id,
                        "download_started_lock": False,
                    },
                    {"result": "OK", "download_result": "download_finished"},
                ),
            ]
        )
        error_logger_mock.assert_not_called()
        storage_location.refresh_from_db()
        self.assertEqual(storage_location.status, StorageLocation.STATUS_DONE)

    @patch("resolwe.flow.managers.listener.consumer.send_event")
    @patch("resolwe.flow.managers.listener.logger.error")
    @patch("resolwe.flow.managers.listener.ExecutorListener._send_reply")
    @patch("resolwe.flow.managers.listener.async_to_sync")
    def test_handle_download_started_ok_lock(
        self, async_to_sync_mock, send_reply_mock, error_logger_mock, send_event_mock
    ):
        storage_location = StorageLocation.objects.create(
            file_storage=self.file_storage, connector_name="local"
        )

        obj = {
            "command": ExecutorProtocol.DOWNLOAD_STARTED,
            "data_id": -1,
            "storage_location_id": storage_location.id,
            "download_started_lock": True,
        }
        self.listener.handle_download_started(obj)

        async_to_sync_mock.assert_has_calls(
            [
                call(send_reply_mock),
                call()(
                    {
                        "command": "download_started",
                        "data_id": -1,
                        "storage_location_id": storage_location.id,
                        "download_started_lock": True,
                    },
                    {"result": "OK", "download_result": "download_started"},
                ),
            ]
        )
        error_logger_mock.assert_not_called()
        storage_location.refresh_from_db()
        self.assertEqual(storage_location.status, StorageLocation.STATUS_UPLOADING)

    @patch("resolwe.flow.managers.listener.consumer.send_event")
    @patch("resolwe.flow.managers.listener.logger.error")
    @patch("resolwe.flow.managers.listener.ExecutorListener._send_reply")
    @patch("resolwe.flow.managers.listener.async_to_sync")
    def test_handle_get_files_to_download_missing_storage_location(
        self, async_to_sync_mock, send_reply_mock, error_logger_mock, send_event_mock
    ):
        obj = {
            "command": ExecutorProtocol.GET_FILES_TO_DOWNLOAD,
            "data_id": -1,
            "storage_location_id": -2,
        }
        self.listener.handle_get_files_to_download(obj)
        async_to_sync_mock.assert_has_calls(
            [
                call(send_reply_mock),
                call()(
                    {
                        "command": "get_files_to_download",
                        "data_id": -1,
                        "storage_location_id": -2,
                    },
                    {"result": "ER"},
                ),
                call(send_event_mock),
                call()(
                    {
                        "command": "abort_data",
                        "data_id": -1,
                        "communicate_kwargs": {
                            "executor": "resolwe.flow.executors.local"
                        },
                    }
                ),
            ]
        )
        error_logger_mock.assert_called_once_with(
            "StorageLocation object does not exist (handle_get_files_to_download).",
            extra={"storage_location_id": -2},
        )

    @patch("resolwe.flow.managers.listener.consumer.send_event")
    @patch("resolwe.flow.managers.listener.logger.error")
    @patch("resolwe.flow.managers.listener.ExecutorListener._send_reply")
    @patch("resolwe.flow.managers.listener.async_to_sync")
    def test_handle_get_files_to_download(
        self, async_to_sync_mock, send_reply_mock, error_logger_mock, send_event_mock
    ):
        obj = {
            "command": ExecutorProtocol.GET_FILES_TO_DOWNLOAD,
            "data_id": -1,
            "storage_location_id": self.storage_location.id,
        }
        self.listener.handle_get_files_to_download(obj)
        async_to_sync_mock.assert_has_calls(
            [
                call(send_reply_mock),
                call()(
                    {
                        "command": "get_files_to_download",
                        "data_id": -1,
                        "storage_location_id": self.storage_location.id,
                    },
                    {
                        "result": "OK",
                        "referenced_files": [
                            {
                                "id": self.path.id,
                                "path": "test.me",
                                "size": -1,
                                "md5": "md5",
                                "crc32c": "crc",
                                "awss3etag": "aws",
                            }
                        ],
                    },
                ),
            ]
        )

    @patch("resolwe.flow.managers.listener.consumer.send_event")
    @patch("resolwe.flow.managers.listener.logger.error")
    @patch("resolwe.flow.managers.listener.ExecutorListener._send_reply")
    @patch("resolwe.flow.managers.listener.async_to_sync")
    def test_handle_get_referenced_files_missing_data(
        self, async_to_sync_mock, send_reply_mock, error_logger_mock, send_event_mock
    ):
        obj = {
            "command": ExecutorProtocol.GET_REFERENCED_FILES,
            "data_id": -1,
        }
        self.listener.handle_get_referenced_files(obj)
        async_to_sync_mock.assert_has_calls(
            [
                call(send_reply_mock),
                call()(
                    {"command": "get_referenced_files", "data_id": -1}, {"result": "ER"}
                ),
                call(send_event_mock),
                call()(
                    {
                        "command": "abort_data",
                        "data_id": -1,
                        "communicate_kwargs": {
                            "executor": "resolwe.flow.executors.local"
                        },
                    }
                ),
            ]
        )
        error_logger_mock.assert_called_once_with(
            "Data object does not exist (handle_get_referenced_files).",
            extra={"data_id": -1},
        )

    @patch("resolwe.flow.managers.listener.consumer.send_event")
    @patch("resolwe.flow.managers.listener.logger.error")
    @patch("resolwe.flow.managers.listener.ExecutorListener._send_reply")
    @patch("resolwe.flow.managers.listener.async_to_sync")
    def test_handle_get_referenced_files(
        self, async_to_sync_mock, send_reply_mock, error_logger_mock, send_event_mock
    ):
        obj = {
            "command": ExecutorProtocol.GET_REFERENCED_FILES,
            "data_id": 1,
        }
        storage_location = StorageLocation.objects.create(
            file_storage=self.file_storage,
            connector_name="local",
            status=StorageLocation.STATUS_DONE,
            url=str(self.file_storage.id),
        )
        path = Path(storage_location.get_path(filename="output.txt"))
        path.parent.mkdir(exist_ok=True, parents=True)
        path.touch()
        data = Data.objects.get(id=1)
        data.process.output_schema = [{"name": "output_file", "type": "basic:file:"}]
        data.process.save()
        data.output = {"output_file": {"file": "output.txt"}}
        data.save()

        self.listener.handle_get_referenced_files(obj)
        async_to_sync_mock.assert_has_calls(
            [
                call(send_reply_mock),
                call()(
                    {"command": "get_referenced_files", "data_id": 1},
                    {
                        "result": "OK",
                        "referenced_files": [
                            "jsonout.txt",
                            "stderr.txt",
                            "stdout.txt",
                            "output.txt",
                        ],
                    },
                ),
            ]
        )
        error_logger_mock.assert_not_called()

    @patch("resolwe.flow.managers.listener.consumer.send_event")
    @patch("resolwe.flow.managers.listener.logger.error")
    @patch("resolwe.flow.managers.listener.ExecutorListener._send_reply")
    @patch("resolwe.flow.managers.listener.async_to_sync")
    def test_handle_missing_data_locations_missing_data(
        self, async_to_sync_mock, send_reply_mock, error_logger_mock, send_event_mock
    ):
        obj = {
            "command": ExecutorProtocol.MISSING_DATA_LOCATIONS,
            "data_id": -1,
        }
        self.listener.handle_missing_data_locations(obj)
        async_to_sync_mock.assert_has_calls(
            [
                call(send_reply_mock),
                call()(
                    {"command": "missing_data_locations", "data_id": -1},
                    {"result": "ER"},
                ),
                call(send_event_mock),
                call()(
                    {
                        "command": "abort_data",
                        "data_id": -1,
                        "communicate_kwargs": {
                            "executor": "resolwe.flow.executors.local"
                        },
                    }
                ),
            ]
        )
        error_logger_mock.assert_called_once_with(
            "Data object does not exist (handle_missing_data_locations).",
            extra={"data_id": -1},
        )

    @patch("resolwe.flow.managers.listener.consumer.send_event")
    @patch("resolwe.flow.managers.listener.logger.error")
    @patch("resolwe.flow.managers.listener.ExecutorListener._send_reply")
    @patch("resolwe.flow.managers.listener.async_to_sync")
    def test_handle_missing_data_locations_missing_storage_location(
        self, async_to_sync_mock, send_reply_mock, error_logger_mock, send_event_mock
    ):
        obj = {
            "command": ExecutorProtocol.MISSING_DATA_LOCATIONS,
            "data_id": 1,
        }
        parent = Data.objects.get(id=2)
        child = Data.objects.get(id=1)
        DataDependency.objects.create(
            parent=parent, child=child, kind=DataDependency.KIND_IO
        )
        self.listener.handle_missing_data_locations(obj)
        async_to_sync_mock.assert_has_calls(
            [
                call(send_reply_mock),
                call()(
                    {"command": "missing_data_locations", "data_id": 1},
                    {"result": "ER"},
                ),
                call(send_event_mock),
                call()(
                    {
                        "command": "abort_data",
                        "data_id": 1,
                        "communicate_kwargs": {
                            "executor": "resolwe.flow.executors.local"
                        },
                    }
                ),
            ]
        )
        error_logger_mock.assert_called_once_with(
            "No storage location exists (handle_get_missing_data_locations).",
            extra={"data_id": 1, "file_storage_id": 2},
        )
        self.assertEqual(StorageLocation.all_objects.count(), 1)

    @patch("resolwe.flow.managers.listener.consumer.send_event")
    @patch("resolwe.flow.managers.listener.logger.error")
    @patch("resolwe.flow.managers.listener.ExecutorListener._send_reply")
    @patch("resolwe.flow.managers.listener.async_to_sync")
    def test_handle_missing_data_locations_none(
        self, async_to_sync_mock, send_reply_mock, error_logger_mock, send_event_mock
    ):
        obj = {
            "command": ExecutorProtocol.MISSING_DATA_LOCATIONS,
            "data_id": 1,
        }
        parent = Data.objects.get(id=2)
        child = Data.objects.get(id=1)
        DataDependency.objects.create(
            parent=parent, child=child, kind=DataDependency.KIND_IO
        )
        StorageLocation.objects.create(
            file_storage=parent.location,
            connector_name="local",
            status=StorageLocation.STATUS_DONE,
            url="url",
        )
        self.listener.handle_missing_data_locations(obj)
        async_to_sync_mock.assert_has_calls(
            [
                call(send_reply_mock),
                call()(
                    {"command": "missing_data_locations", "data_id": 1},
                    {"result": "OK", "storage_data_locations": []},
                ),
            ]
        )
        error_logger_mock.assert_not_called()
        self.assertEqual(StorageLocation.all_objects.count(), 2)

    @patch("resolwe.flow.managers.listener.consumer.send_event")
    @patch("resolwe.flow.managers.listener.logger.error")
    @patch("resolwe.flow.managers.listener.ExecutorListener._send_reply")
    @patch("resolwe.flow.managers.listener.async_to_sync")
    def test_handle_missing_data_locations(
        self, async_to_sync_mock, send_reply_mock, error_logger_mock, send_event_mock
    ):
        obj = {
            "command": ExecutorProtocol.MISSING_DATA_LOCATIONS,
            "data_id": 1,
        }
        parent = Data.objects.get(id=2)
        child = Data.objects.get(id=1)
        DataDependency.objects.create(
            parent=parent, child=child, kind=DataDependency.KIND_IO
        )
        storage_location = StorageLocation.objects.create(
            file_storage=parent.location,
            connector_name="not_local",
            status=StorageLocation.STATUS_DONE,
            url="url",
        )
        self.listener.handle_missing_data_locations(obj)
        self.assertEqual(StorageLocation.all_objects.count(), 3)
        created = StorageLocation.all_objects.last()
        async_to_sync_mock.assert_has_calls(
            [
                call(send_reply_mock),
                call()(
                    {"command": "missing_data_locations", "data_id": child.id},
                    {
                        "result": "OK",
                        "storage_data_locations": [
                            {
                                "connector_name": "not_local",
                                "url": "url",
                                "data_id": child.id,
                                "to_storage_location_id": created.id,
                                "from_storage_location_id": storage_location.id,
                            }
                        ],
                    },
                ),
            ]
        )
        error_logger_mock.assert_not_called()



"""Process runtime."""
import json
import logging
import os
import sys
import urllib

import resolwe_runtime_utils

from .descriptor import ProcessDescriptor
from .fields import Field, RelationDescriptor

try:
    from plumbum import local as Cmd

    # Log plumbum commands to standard output.

    formatter = logging.Formatter(
        "%(asctime)s - %(levelname)s - %(name)s[%(process)s]: %(message)s"
    )

    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(formatter)

    plumbum_logger = logging.getLogger("plumbum.local")
    plumbum_logger.setLevel(logging.DEBUG)
    plumbum_logger.addHandler(handler)


except ImportError:
    # Generate a dummy class that defers the ImportError to a point where
    # the package actually tries to be used.
    class MissingPlumbumCmd:
        """Missing plumbum command dummy class."""

        def __getitem__(self, cmd):
            """Raise ImportError exception."""
            raise ImportError("Missing 'plumbum' Python package.")

    Cmd = MissingPlumbumCmd()

# Inputs class name.
PROCESS_INPUTS_NAME = "Input"
# Outputs class name.
PROCESS_OUTPUTS_NAME = "Output"


class _IoBase:
    """Base class for input/output values."""

    frozen = False

    def __init__(self, fields, **kwargs):
        """Construct an I/O object."""
        super().__setattr__("_data", {})
        super().__setattr__("_fields", fields)
        self._data.update(kwargs)

    def freeze(self):
        """Freeze the object, preventing further mutation."""
        super().__setattr__("frozen", True)
        return self

    def items(self):
        """Return iterator over I/O items."""
        return self._data.items()

    def __getattr__(self, name):
        return self._data.get(name, None)

    def __setattr__(self, name, value):
        if self.frozen:
            raise ValueError("cannot mutate frozen object")

        if name not in self._fields:
            raise ValueError("invalid field name '{}'".format(name))

        value = self._fields[name].clean(value)
        self._data[name] = value

    def __delattr__(self, name):
        if self.frozen:
            raise ValueError("cannot mutate frozen object")

        try:
            del self._data[name]
        except KeyError:
            raise AttributeError(name)

    def __repr__(self):
        return "<{} {}>".format(
            self.__class__.__name__,
            ", ".join(
                [
                    "{}={}".format(field, repr(value))
                    for field, value in self._data.items()
                ]
            ),
        )


class Inputs(_IoBase):
    """Process inputs."""

    def __getattr__(self, name):
        """Get input field.

        Check that the input exists.
        """
        if name not in self._fields:
            raise AttributeError("Inputs have no field {}".format(name))

        return super().__getattr__(name)


class Outputs(_IoBase):
    """Process outputs."""

    def __getattr__(self, name):
        """Get output field.

        Check that the output exists.
        """
        if name not in self._fields:
            raise AttributeError("Outputs have no field {}".format(name))

        return super().__getattr__(name)

    def __setattr__(self, name, value):
        """Store attribute value and save output."""
        super().__setattr__(name, value)

        output_value = self._fields[name].to_output(self._data[name])
        print(json.dumps(output_value))


class ProcessMeta(type):
    """Metaclass for processes during runtime."""

    def __new__(mcs, name, bases, namespace, **kwargs):
        """Create new process class."""
        if namespace["__module__"] == "resolwe.process.runtime":
            return type.__new__(mcs, name, bases, namespace)

        # Generate a runtime version of the process descriptor.
        meta = ProcessDescriptor()
        for meta_name in dir(meta.metadata):
            if meta_name.startswith("_"):
                continue

            setattr(meta.metadata, meta_name, namespace.pop(meta_name, None))

        for nsp, fields in (
            (PROCESS_INPUTS_NAME, meta.inputs),
            (PROCESS_OUTPUTS_NAME, meta.outputs),
        ):
            inputs = namespace.pop(nsp, {})
            for field_name in dir(inputs):
                if field_name.startswith("_"):
                    continue

                field = getattr(inputs, field_name)
                if not isinstance(field, Field):
                    continue
                field.contribute_to_class(meta, fields, field_name)

        meta.validate()

        result = type.__new__(mcs, name, bases, namespace)
        result._meta = meta
        return result


class Process(metaclass=ProcessMeta):
    """Resolwe process."""

    def __init__(self):
        """Construct a new process instance."""
        self.logger = logging.getLogger(self.__class__.__name__)

    def run(self, inputs, outputs):
        """Process entry point."""
        raise NotImplementedError

    def run_process(self, slug, inputs):
        """Run a new process from a running process."""

        def export_files(value):
            """Export input files of spawned process."""
            if isinstance(value, str) and os.path.isfile(value):
                # TODO: Use the protocol to export files and get the
                # process schema to check field type.
                print("export {}".format(value))
            elif isinstance(value, dict):
                for item in value.values():
                    export_files(item)
            elif isinstance(value, list):
                for item in value:
                    export_files(item)

        export_files(inputs)
        print(
            "run {}".format(
                json.dumps({"process": slug, "input": inputs}, separators=(",", ":"))
            )
        )

    def progress(self, progress):
        """Report process progress.

        :param progress: A float between 0 and 1 denoting the progress
        """
        report = resolwe_runtime_utils.progress(progress)
        # TODO: Use the protocol to report progress.
        print(report)

    def info(self, *args):
        """Log informational message."""
        report = resolwe_runtime_utils.info(" ".join([str(x) for x in args]))
        # TODO: Use the protocol to report progress.
        print(report)

    def warning(self, *args):
        """Log warning message."""
        report = resolwe_runtime_utils.warning(" ".join([str(x) for x in args]))
        # TODO: Use the protocol to report progress.
        print(report)

    def error(self, *args):
        """Log error message."""
        report = resolwe_runtime_utils.error(" ".join([str(x) for x in args]))
        # TODO: Use the protocol to report progress.
        print(report)
        sys.exit(1)

    def get_data_id_by_slug(self, slug):
        """Find data object ID for given slug.

        This method queries the Resolwe API and requires network access.
        """
        resolwe_host = os.environ.get("RESOLWE_HOST_URL")
        url = urllib.parse.urljoin(
            resolwe_host, "/api/data?slug={}&fields=id".format(slug)
        )

        with urllib.request.urlopen(url, timeout=60) as f:
            data = json.loads(f.read().decode("utf-8"))

        if len(data) == 1:
            return data[0]["id"]
        elif not data:
            raise ValueError("Data not found for slug {}".format(slug))
        else:
            raise ValueError(
                "More than one data object returned for slug {}".format(slug)
            )

    def update_entity_descriptor(self, annotations):
        """Update entity descriptor from dictionary of annotations."""
        for key, value in annotations.items():
            print(resolwe_runtime_utils.annotate_entity(key, value))

    @property
    def requirements(self):
        """Process requirements."""

        class dotdict(dict):
            """Dot notation access to dictionary attributes."""

            def __getattr__(self, attr):
                value = self.get(attr)
                return dotdict(value) if isinstance(value, dict) else value

        return dotdict(self._meta.metadata.requirements)

    @property
    def name(self):
        """Get Data name."""
        return self._meta.name

    @name.setter
    def name(self, payload):
        """Set Data name."""
        raise NotImplementedError(
            "Setting Data name from within process is not supported."
        )

    @property
    def relations(self):
        """Process relations."""
        return [RelationDescriptor.from_dict(data) for data in self._meta.relations]

    def start(self, inputs):
        """Start the process.

        :param inputs: An instance of `Inputs` describing the process inputs
        :return: An instance of `Outputs` describing the process outputs
        """
        self.logger.info("Process is starting")

        outputs = Outputs(self._meta.outputs)

        self.logger.info("Process is running")
        try:
            self.run(inputs.freeze(), outputs)
            return outputs.freeze()
        except Exception as error:
            self.logger.exception("Exception while running process")
            print(resolwe_runtime_utils.error(str(error)))
            raise
        except:  # noqa
            self.logger.exception("Exception while running process")
            print(resolwe_runtime_utils.error("Exception while running process"))
            raise
        finally:
            self.logger.info("Process has finished")



# Generated by Django 2.2.9 on 2020-04-17 09:53

from django.db import migrations

from resolwe.flow.models.base import delete_chunked
from resolwe.flow.utils.iterators import iterate_fields


def delete_unreferenced_storages(apps, schema_editor):
    Data = apps.get_model("flow", "Data")
    Storage = apps.get_model("flow", "Storage")

    for data in Data.objects.iterator():
        referenced_storages = []
        for field_schema, fields in iterate_fields(
            data.output, data.process.output_schema
        ):
            if field_schema["type"].startswith("basic:json:"):
                name = field_schema["name"]
                referenced_storages.append(fields[name])

        data.storages.remove(*data.storages.exclude(pk__in=referenced_storages))

    delete_chunked(Storage.objects.filter(data__isnull=True), chunk_size=100)


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0044_datalocation_to_filestorage"),
    ]

    operations = [
        migrations.RunPython(delete_unreferenced_storages),
    ]



# pylint: disable=missing-docstring
import os
import shutil
import tempfile
from unittest.mock import MagicMock

from django.contrib.auth import get_user_model
from django.test import override_settings
from django.utils.crypto import get_random_string

from resolwe.flow.executors.collect import collect_and_purge
from resolwe.flow.models import DescriptorSchema, Process
from resolwe.flow.models.utils import referenced_files
from resolwe.test import ProcessTestCase


class PurgeTestFieldsMixin:
    """
    This class contains tests, which validate each field individually. It is used
    to perform basic unit tests (on simulated data, without running any processors)
    as well as more complex end-to-end tests (by running processors). Having them
    in this mixin makes sure that both, unit and e2e tests, are run for each field.
    """

    def test_basic_file(self):
        self.assertFieldWorks(
            "basic:file:",
            field_value={"file": "but_not_this"},
            script_setup="touch remove_this but_not_this",
            script_save="re-save-file sample but_not_this",
            removed=["remove_this"],
            not_removed=["but_not_this"],
        )

    def test_basic_file_specialization(self):
        self.assertFieldWorks(
            "basic:file:html:",
            field_value={"file": "but_not_this"},
            script_setup="touch remove_this but_not_this",
            script_save="re-save-file sample but_not_this",
            removed=["remove_this"],
            not_removed=["but_not_this"],
        )

    def test_basic_file_list(self):
        self.assertFieldWorks(
            "list:basic:file:",
            field_value=[{"file": "but_not_this"}, {"file": "and_not_this"}],
            script_setup="touch remove_this but_not_this and_not_this",
            script_save="re-save-file-list sample but_not_this and_not_this",
            removed=["remove_this"],
            not_removed=["but_not_this", "and_not_this"],
        )

    def test_basic_dir(self):
        self.assertFieldWorks(
            "basic:dir:",
            field_value={"dir": "but_not_this"},
            script_setup="""
mkdir remove_this but_not_this
touch remove_this/a remove_this/b remove_this/c
touch but_not_this/a but_not_this/b but_not_this/c
""",
            script_save="re-save-dir sample but_not_this",
            removed=["remove_this/", "remove_this/a", "remove_this/b", "remove_this/c"],
            not_removed=["but_not_this/a", "but_not_this/b", "but_not_this/c"],
        )

    def test_basic_dir_list(self):
        self.assertFieldWorks(
            "list:basic:dir:",
            field_value=[{"dir": "but_not_this"}, {"dir": "and_not_this"}],
            script_setup="""
mkdir remove_this but_not_this and_not_this
touch remove_this/a remove_this/b remove_this/c
touch but_not_this/a but_not_this/b but_not_this/c
touch and_not_this/a and_not_this/b and_not_this/c
""",
            script_save="re-save-dir-list sample but_not_this and_not_this",
            removed=["remove_this/", "remove_this/a", "remove_this/b", "remove_this/c"],
            not_removed=[
                "but_not_this/a",
                "but_not_this/b",
                "but_not_this/c",
                "and_not_this/a",
                "and_not_this/b",
                "and_not_this/c",
            ],
        )


@override_settings(TEST_PROCESS_REQUIRE_TAGS=False)  # Test uses dynamic processes.
class PurgeE2ETest(PurgeTestFieldsMixin, ProcessTestCase):
    def create_and_run_processor(self, processor, **kwargs):
        processor_slug = get_random_string(6)
        Process.objects.create(
            slug=processor_slug,
            name="Test Purge Process",
            contributor=self.admin,
            type="data:test",
            version=1,
            **processor,
        )
        data = self.run_process(processor_slug, **kwargs)
        return data

    def assertFilesRemoved(self, data, *files):
        for name in files:
            path = data.location.get_path(filename=name)
            self.assertFalse(os.path.isfile(path), msg=path)

    def assertFilesNotRemoved(self, data, *files):
        for name in files:
            path = data.location.get_path(filename=name)
            self.assertTrue(os.path.isfile(path), msg=path)

    def test_complex_processor(self):
        data = self.create_and_run_processor(
            processor=dict(
                input_schema=[],
                output_schema=[
                    {
                        "name": "sample_file",
                        "label": "Sample output file",
                        "type": "basic:file:",
                    },
                    {
                        "name": "another_sample_file",
                        "label": "Another sample output file",
                        "type": "basic:file:",
                    },
                    {
                        "name": "sample_dir",
                        "label": "Sample output directory",
                        "type": "basic:dir:",
                    },
                    {
                        "name": "sample_file_list",
                        "label": "Sample list of output files",
                        "type": "list:basic:file:",
                    },
                    {
                        "name": "sample_dir_list",
                        "label": "Sample list of output directories",
                        "type": "list:basic:dir:",
                    },
                ],
                run={
                    "language": "bash",
                    "program": """
touch these files should be removed
touch this_file_should_stay
mkdir -p should_stay/
touch should_stay/file
mkdir -p directory/should/be
touch directory/should/be/removed
touch directory/should/be/removed2
mkdir -p stay/directory
touch stay/directory/file1
touch stay/directory/file2
touch entry1 entry2 entry3
mkdir dir1 dir2 dir3
touch dir1/a dir2/b dir3/c dir3/d
touch ref1 ref2 ref3 ref4
mkdir refs
touch refs/a refs/b

re-save-file another_sample_file should_stay/file
re-save-file sample_file this_file_should_stay ref3
re-save-dir sample_dir stay ref4
re-save-file-list sample_file_list entry1 entry2:ref1 entry3:refs
re-save-dir-list sample_dir_list dir1:ref2 dir2 dir3
""",
                },
            )
        )

        self.assertFilesRemoved(
            data, "these", "files", "should", "be", "removed", "directory"
        )
        self.assertFilesNotRemoved(
            data,
            "this_file_should_stay",
            "stay/directory/file1",
            "stay/directory/file2",
            "entry1",
            "entry2",
            "entry3",
            "dir1/a",
            "dir2/b",
            "dir3/c",
            "dir3/d",
            "ref1",
            "ref2",
            "ref3",
            "ref4",
            "refs/a",
            "refs/b",
            "should_stay/file",
        )

    def assertFieldWorks(
        self, field_type, field_value, script_setup, script_save, removed, not_removed
    ):
        """
        Checks that a field is handled correctly when running a processor, which
        uses the field.
        """

        field_schema = {"name": "sample", "label": "Sample output", "type": field_type}

        # Test output.
        data = self.create_and_run_processor(
            processor=dict(
                input_schema=[],
                output_schema=[field_schema],
                run={"language": "bash", "program": script_setup + "\n" + script_save},
            )
        )

        self.assertFilesRemoved(data, *removed)
        self.assertFilesNotRemoved(data, *not_removed)

        # Test descriptor.
        descriptor_schema = DescriptorSchema.objects.create(
            slug=get_random_string(6), contributor=self.admin, schema=[field_schema]
        )
        data = self.create_and_run_processor(
            processor=dict(
                input_schema=[],
                output_schema=[],
                run={"language": "bash", "program": script_setup},
            ),
            descriptor_schema=descriptor_schema,
            descriptor={"sample": field_value},
        )

        self.assertFilesRemoved(data, *removed)
        self.assertFilesNotRemoved(data, *not_removed)


class PurgeUnitTest(PurgeTestFieldsMixin, ProcessTestCase):
    def assertFieldWorks(
        self, field_type, field_value, script_setup, script_save, removed, not_removed
    ):
        """
        Checks that a field is handled correctly by `get_purge_files` under a
        simulated Data object.
        """

        field_schema = {"name": "sample", "label": "Sample output", "type": field_type}

        # Test simulated operation.
        simulated_root = tempfile.mkdtemp()
        try:
            for filename in removed + not_removed:
                directory, basename = os.path.split(filename)
                if directory:
                    try:
                        os.makedirs(os.path.join(simulated_root, directory))
                    except OSError:
                        pass

                if basename:
                    with open(os.path.join(simulated_root, filename), "w"):
                        pass

            data_mock = MagicMock(
                output={"sample": field_value},
                process=MagicMock(output_schema=[field_schema]),
                descriptor={},
                descriptor_schema=[],
            )
            refs = referenced_files(data_mock)
            collected, deleted = collect_and_purge(simulated_root, refs)

            def strip_slash(filename):
                return filename[:-1] if filename[-1] == "/" else filename

            collected = [strip_slash(e) for e in collected]
            deleted = [strip_slash(e) for e in deleted]

            for filename in not_removed:
                self.assertIn(
                    strip_slash(filename),
                    collected,
                )

            for filename in removed:
                filename = strip_slash(filename)
                self.assertNotIn(filename, collected)
                self.assertIn(filename, deleted)
                deleted.remove(filename)

            # Ensure that nothing more is removed.
            self.assertEqual(len(deleted), 0)
        finally:
            shutil.rmtree(simulated_root)

    def create_test_file(self, location, filename):
        """
        Creates an empty file in the proper directory, so that a file can be used
        as a reference in the Data object's output/descriptor.
        """
        root = location.get_path()
        try:
            os.makedirs(root)
            self.test_files.add(root)
        except OSError:
            pass

        filename = location.get_path(filename=filename)
        with open(filename, "w"):
            self.test_files.add(filename)

    def setUp(self):
        super().setUp()

        self.test_files = set()

        self.user = get_user_model().objects.create(username="test_user")
        processor = Process.objects.create(
            name="Test process",
            contributor=self.user,
            output_schema=[{"name": "sample", "type": "basic:file:"}],
        )
        self.data = {
            "name": "Test data",
            "contributor": self.user,
            "process": processor,
        }

    def tearDown(self):
        for filename in self.test_files:
            if os.path.isfile(filename):
                os.remove(filename)
            elif os.path.isdir(filename):
                shutil.rmtree(filename)

        super().tearDown()



"""Transfer missing files to executor."""
import logging
import os
import sys
import time

from .global_settings import DATA_META
from .manager_commands import send_manager_command
from .protocol import ExecutorProtocol

logger = logging.getLogger(__name__)

# Make sure sphinx can import this module.
try:
    from .connectors import Transfer, connectors
    from .connectors.exceptions import DataTransferError
except ImportError:
    logger.exception("Unable to import 'connectors' module")


DOWNLOAD_WAITING_TIMEOUT = 60  # in seconds
RETRIES = 5


async def _notify_abort(reason):
    """Notify listener that worker is aborting processing due to error."""
    await send_manager_command(
        ExecutorProtocol.UPDATE,
        extra_fields={
            ExecutorProtocol.UPDATE_CHANGESET: {
                "process_error": [reason],
                "status": DATA_META["STATUS_ERROR"],
            }
        },
    )
    await send_manager_command(ExecutorProtocol.ABORT, expect_reply=False)


async def transfer_data():
    """Transfer missing data, terminate script on failure."""
    try:
        abort_reason = None
        if not await _transfer_data():
            abort_reason = "Error while transfering data."
    except RuntimeError:
        abort_reason = "Communication error: data object or its location is missing."
    finally:
        if abort_reason:
            await _notify_abort(abort_reason)
            sys.exit(1)


async def _transfer_data():
    """Fetch missing data.

    Get a list of missing storage locations from the manager fetch
    data using appropriate storage connectors.
    """
    result = await send_manager_command(ExecutorProtocol.MISSING_DATA_LOCATIONS)
    data_to_transfer = result[ExecutorProtocol.STORAGE_DATA_LOCATIONS]

    to_connector = connectors["local"]
    base_path = to_connector.base_path
    data_downloading = []

    # Notify manager to change status of the data object.
    if data_to_transfer:
        await send_manager_command(
            ExecutorProtocol.UPDATE,
            extra_fields={
                ExecutorProtocol.UPDATE_CHANGESET: {
                    "status": DATA_META["STATUS_PREPARING"]
                }
            },
        )
    # Start downloading data.
    while data_to_transfer or data_downloading:
        if data_to_transfer:
            missing_data = data_to_transfer.pop()
            storage_location_id = missing_data["to_storage_location_id"]
            logger.debug(
                "Requesting download lock for storage location with id {}".format(
                    storage_location_id
                )
            )
            reply = await send_manager_command(
                ExecutorProtocol.DOWNLOAD_STARTED,
                extra_fields={
                    ExecutorProtocol.STORAGE_LOCATION_ID: storage_location_id,
                    ExecutorProtocol.DOWNLOAD_STARTED_LOCK: True,
                },
            )
            result = reply[ExecutorProtocol.DOWNLOAD_RESULT]
            # There are three options:
            # - download is already finished
            # - we can start downloading
            # - download is already in progress
            # The first one needs no special processing.
            if result == ExecutorProtocol.DOWNLOAD_STARTED:
                # We are good to go for download
                logger.debug(
                    "Download lock for storage location with id {} obtained".format(
                        storage_location_id
                    )
                )
                dest_dir = os.path.join(base_path, missing_data["url"])
                os.makedirs(dest_dir, exist_ok=True)
                # Download data will abort or finish the download
                if not await download_data(missing_data):
                    logger.error(
                        "Download for data with id {} aborted".format(
                            missing_data["data_id"]
                        )
                    )
                    return False
            elif result == ExecutorProtocol.DOWNLOAD_IN_PROGRESS:
                # Somebody else is downloading the data, wait for it
                logger.debug(
                    "Download for storage location with id {} already in progress".format(
                        storage_location_id
                    )
                )
                data_downloading.append(missing_data)

        elif data_downloading:
            logger.debug("Waiting for downloads {} to finish".format(data_downloading))
            time.sleep(DOWNLOAD_WAITING_TIMEOUT)
            data_downloading_new = []
            for download in data_downloading:
                storage_location_id = download["to_storage_location_id"]
                # Query for the status of the download
                reply = await send_manager_command(
                    ExecutorProtocol.DOWNLOAD_STARTED,
                    extra_fields={
                        ExecutorProtocol.STORAGE_LOCATION_ID: storage_location_id,
                        ExecutorProtocol.DOWNLOAD_STARTED_LOCK: False,
                    },
                )
                result = reply[ExecutorProtocol.DOWNLOAD_RESULT]
                # We have 3 options:
                # - download was aborted, we can restart it
                # - download is still in progress
                # - download finished
                if result == ExecutorProtocol.DOWNLOAD_STARTED:
                    # We can start the download
                    data_to_transfer.append(download)
                    logger.debug("Remote download {} was aborted".format(download))
                elif result == ExecutorProtocol.DOWNLOAD_IN_PROGRESS:
                    # Download is still in progress, wait
                    data_downloading_new.append(download)
                    logger.debug("Download {} is still in progress".format(download))
            data_downloading = data_downloading_new
    return True


async def download_data(missing_data: dict) -> bool:
    """Download missing data for one data object.

    :returns: status of the transfer: True for success, False for failure.
    :rtype: bool
    """
    data_id = missing_data["data_id"]
    logger.debug(
        "Starting data transfer for data_id {}".format(missing_data["data_id"])
    )
    objects = None
    for retry in range(1, RETRIES + 1):
        try:
            to_connector = connectors["local"]
            from_connector = connectors[missing_data["connector_name"]]
            from_storage_location_id = missing_data["from_storage_location_id"]
            to_storage_location_id = missing_data["to_storage_location_id"]

            if objects is None:
                response = await send_manager_command(
                    ExecutorProtocol.GET_FILES_TO_DOWNLOAD,
                    extra_fields={
                        ExecutorProtocol.STORAGE_LOCATION_ID: from_storage_location_id,
                    },
                )
                objects = response[ExecutorProtocol.REFERENCED_FILES]

            t = Transfer(from_connector, to_connector)
            t.transfer_objects(missing_data["url"], objects)
            await send_manager_command(
                ExecutorProtocol.DOWNLOAD_FINISHED,
                extra_fields={
                    ExecutorProtocol.STORAGE_LOCATION_ID: to_storage_location_id
                },
            )
            return True
        except DataTransferError:
            logger.exception(
                "Data transfer error downloading data with id {}, retry {} out of".format(
                    data_id, retry, RETRIES
                )
            )
        except Exception:
            logger.exception(
                "Unknown error downloading data with id {}, retry {} out of".format(
                    data_id, retry, RETRIES
                )
            )
    # None od the retries has been successfull, abort the download.
    await send_manager_command(
        ExecutorProtocol.DOWNLOAD_ABORTED,
        expect_reply=False,
        extra_fields={ExecutorProtocol.STORAGE_LOCATION_ID: to_storage_location_id},
    )
    return False



# pylint: disable=missing-docstring
import asyncio
import json
import sys
from functools import partial
from unittest.mock import MagicMock, call, patch

from resolwe.flow import executors
from resolwe.flow.managers.protocol import ExecutorProtocol
from resolwe.flow.models import Data
from resolwe.storage import connectors
from resolwe.storage.connectors.exceptions import DataTransferError
from resolwe.test import TestCase


async def coroutine(*args, **kwargs):
    """Coroutine that just returns its first argument."""
    return args[0]


async def send(rules, command, extra_fields=None, expect_reply=True):
    """Coroutine used to mock send_manager_command. It accepts the list of
    expected calls and replies. When the actual call deviates from these
    list the message with status ExecutorProtocol.RESULT_ERROR is sent to the
    executor.
    """
    rule_index = rules[0]
    rules[0] += 1
    expected, response = rules[rule_index]
    if [command, extra_fields] != expected:
        return {ExecutorProtocol.RESULT: ExecutorProtocol.RESULT_ERROR}
    else:
        if callable(response):
            return response()
        else:
            return response


async def start(actual_coroutine):
    """
    Start the testing coroutine and wait 1 second for it to complete.
    :raises asyncio.CancelledError when the coroutine fails to finish its work
        in 1 second.
    :returns: the return value of the actual_coroutine.
    :rtype: Any

    """
    try:
        return await asyncio.wait_for(actual_coroutine, 2)
    except asyncio.CancelledError:
        pass


def run_async(coroutine):
    loop = asyncio.new_event_loop()
    return loop.run_until_complete(start(coroutine))


def raise_datatransfererror():
    raise DataTransferError()


class SettingsJSONifier(json.JSONEncoder):
    """Customized JSON encoder, coercing all unknown types into strings.

    Needed due to the class hierarchy coming out of the database,
    which can't be serialized using the vanilla json encoder.
    """

    def default(self, o):
        """Try default; otherwise, coerce the object into a string."""
        try:
            return super().default(o)
        except TypeError:
            return str(o)


MODULES_PATCH = {"resolwe.flow.executors.connectors": connectors, "sphinx": MagicMock()}
TRANSFER = "resolwe.flow.executors.transfer"
TRANSFER_PATCHES = {
    "DATA_META": {
        k: getattr(Data, k)
        for k in dir(Data)
        if k.startswith("STATUS_") and isinstance(getattr(Data, k), str)
    }
}


class BasicTestCase(TestCase):
    RESULT_ERROR = {ExecutorProtocol.RESULT: ExecutorProtocol.RESULT_ERROR}
    RESULT_OK = {ExecutorProtocol.RESULT: ExecutorProtocol.RESULT_OK}
    DOWNLOAD_STARTED = {
        ExecutorProtocol.RESULT: ExecutorProtocol.RESULT_OK,
        ExecutorProtocol.DOWNLOAD_RESULT: ExecutorProtocol.DOWNLOAD_STARTED,
    }
    DOWNLOAD_IN_PROGRESS = {
        ExecutorProtocol.RESULT: ExecutorProtocol.RESULT_OK,
        ExecutorProtocol.DOWNLOAD_RESULT: ExecutorProtocol.DOWNLOAD_IN_PROGRESS,
    }
    DOWNLOAD_FINISHED = {
        ExecutorProtocol.RESULT: ExecutorProtocol.RESULT_OK,
        ExecutorProtocol.DOWNLOAD_RESULT: ExecutorProtocol.DOWNLOAD_FINISHED,
    }
    ACCESS_LOG = {
        ExecutorProtocol.RESULT: ExecutorProtocol.RESULT_OK,
        ExecutorProtocol.STORAGE_ACCESS_LOG_ID: 1,
    }


@patch.dict(sys.modules, MODULES_PATCH)
@patch.multiple(TRANSFER, **TRANSFER_PATCHES)
class TransfersTest(BasicTestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        self.missing_data = [
            {
                "connector_name": "CONNECTOR",
                "url": "URL",
                "data_id": "1",
                "from_storage_location_id": 1,
                "to_storage_location_id": 2,
            }
        ]
        self.DOWNLOAD_STARTED_LOCK = [
            ExecutorProtocol.DOWNLOAD_STARTED,
            {
                ExecutorProtocol.STORAGE_LOCATION_ID: 2,
                ExecutorProtocol.DOWNLOAD_STARTED_LOCK: True,
            },
        ]
        self.DOWNLOAD_STARTED_NO_LOCK = [
            ExecutorProtocol.DOWNLOAD_STARTED,
            {
                ExecutorProtocol.STORAGE_LOCATION_ID: 2,
                ExecutorProtocol.DOWNLOAD_STARTED_LOCK: False,
            },
        ]
        self.MISSING_DATA = [
            ExecutorProtocol.MISSING_DATA_LOCATIONS,
            None,
        ]
        self.MISSING_DATA_RESPONSE = {
            ExecutorProtocol.RESULT: ExecutorProtocol.RESULT_OK,
            ExecutorProtocol.STORAGE_DATA_LOCATIONS: self.missing_data.copy(),
        }

        return super().setUp()

    def _test_workflow(self, send_command, download_command, commands, result):
        with patch.multiple(
            TRANSFER, **self.get_patches(send_command, download_command)
        ):
            result = run_async(executors.transfer._transfer_data())
        self.assertEqual(send_command.call_count, len(commands) - 1)
        expected_args_list = []
        for data, _ in commands[1:]:
            command, extra_fields = data
            if extra_fields is not None:
                expected_args_list.append(call(command, extra_fields=extra_fields))
            else:
                expected_args_list.append(call(command))
        self.assertEqual(expected_args_list, send_command.call_args_list)
        return result

    def get_patches(self, send_command, download_command=None):
        patches = {
            "send_manager_command": send_command,
            "DOWNLOAD_WAITING_TIMEOUT": 0,
        }
        if download_command is not None:
            patches["download_data"] = download_command
        return patches

    def test_no_transfer(self):
        self.MISSING_DATA_RESPONSE = {
            ExecutorProtocol.RESULT: ExecutorProtocol.RESULT_OK,
            ExecutorProtocol.STORAGE_DATA_LOCATIONS: [],
        }
        send_command = MagicMock(
            side_effect=[
                coroutine(self.MISSING_DATA_RESPONSE),
                coroutine(self.RESULT_OK),
            ]
        )
        with patch.multiple(TRANSFER, send_manager_command=send_command):
            run_async(executors.transfer._transfer_data())
        send_command.assert_called_once_with("missing_data_locations")

    def test_transfer_download(self):
        commands = [
            1,
            (self.MISSING_DATA, self.MISSING_DATA_RESPONSE),
            (["update", {"changeset": {"status": "PP"}}], self.RESULT_OK),
            (self.DOWNLOAD_STARTED_LOCK, self.DOWNLOAD_STARTED),
        ]
        send_command = MagicMock(side_effect=partial(send, commands))
        download_command = MagicMock(return_value=coroutine(True))
        result = self._test_workflow(send_command, download_command, commands, True)
        self.assertTrue(result)
        download_command.assert_called_once_with(self.missing_data[0])

    def test_transfer_wait_download(self):
        commands = [
            1,
            (self.MISSING_DATA, self.MISSING_DATA_RESPONSE),
            (["update", {"changeset": {"status": "PP"}}], self.RESULT_OK),
            (self.DOWNLOAD_STARTED_LOCK, self.DOWNLOAD_IN_PROGRESS),
            (self.DOWNLOAD_STARTED_NO_LOCK, self.DOWNLOAD_STARTED),
            (self.DOWNLOAD_STARTED_LOCK, self.DOWNLOAD_STARTED),
        ]
        download_command = MagicMock(return_value=coroutine(True))
        send_command = MagicMock(side_effect=partial(send, commands))
        result = self._test_workflow(send_command, download_command, commands, True)
        self.assertTrue(result)
        download_command.assert_called_once_with(self.missing_data[0])

    def test_transfer_wait_download_long(self):
        commands = [
            1,
            (self.MISSING_DATA, self.MISSING_DATA_RESPONSE),
            (["update", {"changeset": {"status": "PP"}}], self.RESULT_OK),
            (self.DOWNLOAD_STARTED_LOCK, self.DOWNLOAD_IN_PROGRESS),
            (self.DOWNLOAD_STARTED_NO_LOCK, self.DOWNLOAD_STARTED),
            (self.DOWNLOAD_STARTED_LOCK, self.DOWNLOAD_IN_PROGRESS),
            (self.DOWNLOAD_STARTED_NO_LOCK, self.DOWNLOAD_IN_PROGRESS),
            (self.DOWNLOAD_STARTED_NO_LOCK, self.DOWNLOAD_STARTED),
            (self.DOWNLOAD_STARTED_LOCK, self.DOWNLOAD_IN_PROGRESS),
            (self.DOWNLOAD_STARTED_NO_LOCK, self.DOWNLOAD_IN_PROGRESS),
            (self.DOWNLOAD_STARTED_NO_LOCK, self.DOWNLOAD_STARTED),
            (self.DOWNLOAD_STARTED_LOCK, self.DOWNLOAD_STARTED),
        ]
        download_command = MagicMock(return_value=coroutine(True))
        send_command = MagicMock(side_effect=partial(send, commands))
        result = self._test_workflow(send_command, download_command, commands, True)
        self.assertTrue(result)
        download_command.assert_called_once_with(self.missing_data[0])

    def test_transfer_failed(self):
        commands = [
            1,
            (self.MISSING_DATA, self.MISSING_DATA_RESPONSE),
            (["update", {"changeset": {"status": "PP"}}], self.RESULT_OK),
            (self.DOWNLOAD_STARTED_LOCK, self.DOWNLOAD_STARTED),
        ]
        download_command = MagicMock(return_value=coroutine(False))
        send_command = MagicMock(side_effect=partial(send, commands))
        result = self._test_workflow(send_command, download_command, commands, True)
        self.assertFalse(result)
        download_command.assert_called_once_with(self.missing_data[0])

    def test_transfer_downloaded(self):
        commands = [
            1,
            (self.MISSING_DATA, self.MISSING_DATA_RESPONSE),
            (["update", {"changeset": {"status": "PP"}}], self.RESULT_OK),
            (self.DOWNLOAD_STARTED_LOCK, self.DOWNLOAD_FINISHED),
        ]
        download_command = MagicMock(return_value=coroutine(True))
        send_command = MagicMock(side_effect=partial(send, commands))
        result = self._test_workflow(send_command, download_command, commands, True)
        self.assertTrue(result)
        download_command.assert_not_called()

    def test_transfer_wait_downloaded(self):
        commands = [
            1,
            (self.MISSING_DATA, self.MISSING_DATA_RESPONSE),
            (["update", {"changeset": {"status": "PP"}}], self.RESULT_OK),
            (self.DOWNLOAD_STARTED_LOCK, self.DOWNLOAD_IN_PROGRESS),
            (self.DOWNLOAD_STARTED_NO_LOCK, self.DOWNLOAD_FINISHED),
        ]
        download_command = MagicMock(return_value=coroutine(True))
        send_command = MagicMock(side_effect=partial(send, commands))
        result = self._test_workflow(send_command, download_command, commands, True)
        self.assertTrue(result)
        download_command.assert_not_called()

    def test_transfer_downloadmulti_simple(self):
        self.missing_data = [
            {
                "connector_name": "CONNECTOR",
                "url": "URL",
                "data_id": "1",
                "from_storage_location_id": 1,
                "to_storage_location_id": 2,
            },
            {
                "connector_name": "CONNECTOR",
                "url": "URL",
                "data_id": "1",
                "from_storage_location_id": 1,
                "to_storage_location_id": 2,
            },
        ]
        self.MISSING_DATA_RESPONSE = {
            ExecutorProtocol.RESULT: ExecutorProtocol.RESULT_OK,
            ExecutorProtocol.STORAGE_DATA_LOCATIONS: self.missing_data.copy(),
        }
        commands = [
            1,
            (self.MISSING_DATA, self.MISSING_DATA_RESPONSE),
            (["update", {"changeset": {"status": "PP"}}], self.RESULT_OK),
            (self.DOWNLOAD_STARTED_LOCK, self.DOWNLOAD_STARTED),
            (self.DOWNLOAD_STARTED_LOCK, self.DOWNLOAD_STARTED),
        ]
        download_command = MagicMock(side_effect=lambda e: coroutine(True))
        send_command = MagicMock(side_effect=partial(send, commands))
        result = self._test_workflow(send_command, download_command, commands, True)
        self.assertTrue(result)
        self.assertEqual(download_command.call_count, 2)

    def test_transfer_downloadmulti(self):
        self.missing_data = [
            {
                "connector_name": "CONNECTOR",
                "url": "URL",
                "data_id": "1",
                "from_storage_location_id": 1,
                "to_storage_location_id": 2,
            },
            {
                "connector_name": "CONNECTOR",
                "url": "URL",
                "data_id": "1",
                "from_storage_location_id": 1,
                "to_storage_location_id": 2,
            },
        ]
        self.MISSING_DATA_RESPONSE = {
            ExecutorProtocol.RESULT: ExecutorProtocol.RESULT_OK,
            ExecutorProtocol.STORAGE_DATA_LOCATIONS: self.missing_data.copy(),
        }
        commands = [
            1,
            (self.MISSING_DATA, self.MISSING_DATA_RESPONSE),
            (["update", {"changeset": {"status": "PP"}}], self.RESULT_OK),
            (self.DOWNLOAD_STARTED_LOCK, self.DOWNLOAD_IN_PROGRESS),
            (self.DOWNLOAD_STARTED_LOCK, self.DOWNLOAD_STARTED),
            (self.DOWNLOAD_STARTED_NO_LOCK, self.DOWNLOAD_FINISHED),
        ]
        download_command = MagicMock(side_effect=lambda e: coroutine(True))
        send_command = MagicMock(side_effect=partial(send, commands))
        result = self._test_workflow(send_command, download_command, commands, True)
        self.assertTrue(result)
        download_command.assert_called_once()

    def test_transfer_protocol_fail(self):
        async def raise_exception(*args, **kwargs):
            raise RuntimeError("Protocol error")

        commands = [
            1,
            (self.MISSING_DATA, self.MISSING_DATA_RESPONSE),
            (["update", {"changeset": {"status": "PP"}}], self.RESULT_OK),
            (self.DOWNLOAD_STARTED_LOCK, self.DOWNLOAD_STARTED),
        ]
        download_command = MagicMock(return_value=coroutine(True))
        send_command = MagicMock(side_effect=raise_exception)
        with self.assertRaises(RuntimeError):
            self._test_workflow(send_command, download_command, commands, True)
        download_command.assert_not_called()

    def test_outer_command_success(self):
        send_manager_command = MagicMock()
        _transfer_data = MagicMock(return_value=coroutine(True))
        patches = {
            "_transfer_data": _transfer_data,
            "send_manager_command": send_manager_command,
        }
        with patch.multiple(TRANSFER, **patches):
            result = run_async(executors.transfer.transfer_data())
        self.assertIsNone(result)
        _transfer_data.assert_called_once()
        send_manager_command.assert_not_called()

    def test_outer_command_failure(self):
        send_manager_command = MagicMock()
        commands = [
            1,
            (
                [
                    "update",
                    {
                        "changeset": {
                            "process_error": ["Error while transfering data."],
                            "status": Data.STATUS_ERROR,
                        }
                    },
                ],
                self.RESULT_OK,
            ),
            (["abort", {}], self.RESULT_OK),
        ]
        send_manager_command = MagicMock(side_effect=partial(send, commands))

        _transfer_data = MagicMock(return_value=coroutine(False))
        patches = {
            "_transfer_data": _transfer_data,
            "send_manager_command": send_manager_command,
        }
        with patch.multiple(TRANSFER, **patches):
            with self.assertRaises(SystemExit) as context_manager:
                run_async(executors.transfer.transfer_data())
        self.assertEqual(context_manager.exception.code, 1)
        _transfer_data.assert_called_once()
        calls = [
            call(commands[1][0][0], extra_fields=commands[1][0][1]),
            call(commands[2][0][0], expect_reply=False),
        ]
        self.assertEqual(send_manager_command.call_args_list, calls)


@patch.dict(sys.modules, MODULES_PATCH)
@patch.multiple(TRANSFER, **TRANSFER_PATCHES)
class DownloadDataTest(BasicTestCase):
    def setUp(self):
        self.missing_data = {
            "connector_name": "S3",
            "url": "transfer_url",
            "data_id": 1,
            "from_storage_location_id": 1,
            "to_storage_location_id": 2,
        }
        self.COMMAND_DOWNLOAD_FINISHED = [
            ExecutorProtocol.DOWNLOAD_FINISHED,
            {ExecutorProtocol.STORAGE_LOCATION_ID: 2},
        ]
        self.COMMAND_DOWNLOAD_ABORTED = [
            ExecutorProtocol.DOWNLOAD_ABORTED,
            {ExecutorProtocol.STORAGE_LOCATION_ID: 2},
        ]
        self.COMMAND_GET_FILES = [
            ExecutorProtocol.GET_FILES_TO_DOWNLOAD,
            {ExecutorProtocol.STORAGE_LOCATION_ID: 1},
        ]
        self.FILES_LIST = {
            ExecutorProtocol.RESULT: ExecutorProtocol.RESULT_OK,
            ExecutorProtocol.REFERENCED_FILES: ["1", "dir/1"],
        }

        return super().setUp()

    def get_patches(self, send_command, transfer_module):
        patches = {
            "send_manager_command": send_command,
            "Transfer": transfer_module,
            "DataTransferError": DataTransferError,
            "RETRIES": 2,
        }
        return patches

    def _test_workflow_calls(self, send_command, commands):
        self.assertEqual(send_command.call_count, len(commands) - 1)
        expected_args_list = []
        for i in range(1, len(commands)):
            command, extra_fields = commands[i][0]
            kwargs = {"extra_fields": extra_fields}
            if command in (ExecutorProtocol.DOWNLOAD_ABORTED,):
                kwargs["expect_reply"] = False
            expected_args_list.append(call(command, **kwargs))
        self.assertEqual(expected_args_list, send_command.call_args_list)

    def _test_workflow(
        self, send_command, transfer_module, commands, expected_result=True
    ):
        with patch.multiple(
            TRANSFER,
            **self.get_patches(send_command, transfer_module),
        ):
            result = run_async(
                executors.transfer.download_data(self.missing_data.copy())
            )
        self.assertEqual(result, expected_result)
        self._test_workflow_calls(send_command, commands)

    def test_download_simple(self):
        commands = [
            1,
            (self.COMMAND_GET_FILES, self.FILES_LIST),
            (self.COMMAND_DOWNLOAD_FINISHED, self.RESULT_OK),
        ]
        send_command = MagicMock(side_effect=partial(send, commands))

        # Mock the transfer object
        transfer_objects = MagicMock()
        transfer = MagicMock(transfer_objects=transfer_objects)
        transfer_module = MagicMock(return_value=transfer)
        self._test_workflow(send_command, transfer_module, commands)
        transfer_objects.assert_called_once_with("transfer_url", ["1", "dir/1"])

    def test_download_retry(self):
        commands = [
            1,
            (self.COMMAND_GET_FILES, self.FILES_LIST),
            (self.COMMAND_DOWNLOAD_FINISHED, raise_datatransfererror),
            (self.COMMAND_DOWNLOAD_FINISHED, self.RESULT_OK),
        ]
        send_command = MagicMock(side_effect=partial(send, commands))

        # Mock the transfer object
        transfer_objects = MagicMock()
        transfer = MagicMock(transfer_objects=transfer_objects)
        transfer_module = MagicMock(return_value=transfer)
        self._test_workflow(send_command, transfer_module, commands)
        self.assertEqual(
            transfer_objects.call_args_list,
            [call("transfer_url", ["1", "dir/1"])] * 2,
        )

    def test_download_retry_fail(self):
        commands = [
            1,
            (self.COMMAND_GET_FILES, self.FILES_LIST),
            (self.COMMAND_DOWNLOAD_FINISHED, raise_datatransfererror),
            (self.COMMAND_DOWNLOAD_FINISHED, raise_datatransfererror),
            (self.COMMAND_DOWNLOAD_ABORTED, self.RESULT_OK),
        ]
        send_command = MagicMock(side_effect=partial(send, commands))

        # Mock the transfer object
        transfer_objects = MagicMock()
        transfer = MagicMock(transfer_objects=transfer_objects)
        transfer_module = MagicMock(return_value=transfer)

        self._test_workflow(send_command, transfer_module, commands, False)
        self.assertEqual(
            transfer_objects.call_args_list,
            [call("transfer_url", ["1", "dir/1"])] * 2,
        )



from resolwe.process import DataField, Process


class PythonProcessJson(Process):
    """This is a process description."""

    slug = "test-python-process-json"
    name = "Python Process that uses JSON field"
    process_type = "data:python"
    version = "0.1.2"
    requirements = {
        "executor": {
            "docker": {
                "image": "resolwe/base:ubuntu-18.04",
            }
        }
    }

    class Input:
        """Input fields."""

        data = DataField("test", label="My input data")
        data2 = DataField(
            "test", label="My second non-required input data", required=False
        )

    def run(self, inputs, outputs):
        print("Input data:", inputs.data)
        print("Input data:", inputs.data.storage)



# pylint: disable=missing-docstring
import sys
import unittest

from django.contrib.auth.models import AnonymousUser
from django.test import LiveServerTestCase

from guardian.shortcuts import assign_perm

from resolwe.flow.models import Collection, Process, Storage
from resolwe.test import (
    ProcessTestCase,
    check_installed,
    tag_process,
    with_docker_executor,
    with_resolwe_host,
)


class AccessAPIFromExecutorProcessTestCase(ProcessTestCase, LiveServerTestCase):
    def _create_collection(self):
        """Create a test collection that will be accessed via a test process.

        Assign :class:`~django.contrib.auth.models.AnonymousUser`
        ``view_collection`` permission to the created collection.

        :return: created test collection
        :rtype: Collection

        """
        collection = Collection.objects.create(
            name="collection-foo",
            contributor=self.contributor,
        )
        assign_perm("view_collection", AnonymousUser(), collection)
        return collection

    def create_process(self):
        """Create a test process that will query the API for collections.

        Get the list of all collections using ``curl`` to query the Live
        Resolwe host.

        :return: created test process
        :rtype: Process

        """
        process = Process.objects.create(
            name="Test accessing API from process",
            requirements={"expression-engine": "jinja", "resources": {"network": True}},
            contributor=self.contributor,
            type="data:test:api-access",
            input_schema=[],
            output_schema=[
                {
                    "name": "collection-list",
                    "type": "basic:json:",
                    "required": False,
                },
            ],
            run={
                "language": "bash",
                # NOTE: Resolwe Runtime Utilities are not available in the
                # resolwe/test:base Docker image so we need to use plain echo
                # NOTE: curl must be called with --silent argument to not ouput
                # its progress since it can interfere with parsing of JSON on
                # process' standard output
                "program": r"""
echo "{\"collection-list\": $(curl --silent --show-error $RESOLWE_HOST_URL/api/collection)}"
""",
            },
        )
        return process

    def setUp(self):
        """Custom initilization of :class:`~ProcessTestCase`."""
        super().setUp()

        self.process = self.create_process()

    def check_results(self, data):
        """Check if information obtained for collection via API is correct."""
        if "collection-list" not in data.output:
            self.fail(
                "The test process didn't obtain collection list from the live Resolwe host."
                + self._debug_info(data)
            )
        collection_list = Storage.objects.get(data=data).json
        self.assertEqual(len(collection_list), 1)
        self.assertEqual(collection_list[0]["slug"], self.collection.slug)

    @unittest.skipUnless(*check_installed("curl"))
    @with_resolwe_host
    @tag_process("test-accessing-api-from-process")
    def test_access_api_from_local_executor_process(self):
        """Test if a process running via local executor can access API."""
        data = self.run_process(self.process.slug)
        self.check_results(data)

    @unittest.skipUnless(*check_installed("curl"))
    @tag_process("test-accessing-api-from-process")
    def test_access_api_from_local_executor_process_without_decorator(self):
        """Test if a process running via local executor cannot access API."""
        data = self.run_process(self.process.slug)
        with self.assertRaises(AssertionError):
            self.check_results(data)

    @unittest.skipUnless(
        sys.platform.startswith("linux"),
        "Accessing live Resolwe host from a Docker container on non-Linux systems is not possible yet.",
    )
    @with_resolwe_host
    @with_docker_executor
    @tag_process("test-accessing-api-from-process")
    def test_access_api_from_docker_executor_process(self):
        """Test if a process running via Docker executor can access API."""

        # NOTE: pylint gets confused if the above string is inline with the method name
        data = self.run_process(self.process.slug)
        self.check_results(data)

    @unittest.skipUnless(
        sys.platform.startswith("linux"),
        "Accessing live Resolwe host from a Docker container on non-Linux systems is not possible yet.",
    )
    @with_docker_executor
    @with_resolwe_host
    @tag_process("test-accessing-api-from-process")
    def test_access_api_from_docker_executor_process_reverse_decorators(self):
        """Test if a process running via Docker executor can access API."""

        # NOTE: pylint gets confused if the above string is inline with the method name
        data = self.run_process(self.process.slug)
        self.check_results(data)



"""Collect files after worker is finished with processing."""
import logging
import os
import shutil
from pathlib import Path
from typing import List, Set, Tuple

from .protocol import ExecutorProtocol

# Make sure sphinx can import this module.
try:
    from .connectors.utils import get_transfer_object
except ImportError:
    pass


logger = logging.getLogger(__name__)


def collect_and_purge(base_path: str, refs: List[str]) -> Tuple[Set[str], Set[str]]:
    """Collect all files and directories that are listed in the list refs."""

    def collect_parent_dirs(base_path: str, path: str) -> List[str]:
        """Also collect parent directories until base_path."""
        base_path = os.path.join(base_path, "")
        parents = []
        path = os.path.dirname(path)
        while path.startswith(base_path):
            parents.append(os.path.join(os.path.relpath(path, base_path), ""))
            path = os.path.dirname(path)
        return parents

    def get_dir_files(base_dir: str, dir_: str) -> Set[str]:
        """Get a list of all elements in the directory.

        Returned paths are relative with respect to the base_dir.
        """
        collected = set()
        for root, _, files in os.walk(dir_):
            collected.add(os.path.join(os.path.relpath(root, base_dir), ""))
            collected.update(
                [
                    os.path.relpath(os.path.join(root, file_), base_dir)
                    for file_ in files
                ]
            )
        return collected

    collected = set()
    removed = set()

    for ref in refs:
        real_path = os.path.join(base_path, ref)
        if not os.path.exists(real_path):
            continue
        collected.update(collect_parent_dirs(base_path, real_path))
        if os.path.isdir(real_path):
            collected.update(get_dir_files(base_path, real_path))
        else:
            collected.add(ref)

    # Remove unreferenced files.
    for root, dirs, files in os.walk(os.path.join(base_path)):
        for file_ in files:
            real_path = os.path.join(root, file_)
            relative_path = os.path.relpath(real_path, base_path)
            if relative_path not in collected:
                os.remove(real_path)
                removed.add(relative_path)
        # Modify dirs in place in order not to visit removed directories.
        i = 0
        while i < len(dirs):
            dir_ = dirs[i]
            real_path = os.path.join(root, dir_)
            relative_path = os.path.join(os.path.relpath(real_path, base_path), "")
            if relative_path not in collected:
                if os.path.isdir(real_path):
                    removed.update(get_dir_files(base_path, real_path))
                    shutil.rmtree(real_path)
                dirs.pop(i)
            else:
                i = i + 1
    return collected, removed


async def collect_files():
    """Collect files produced by the worker.

    Keep only files that are referenced in the data model.
    """
    # Make file importable from outside executor environment
    from .global_settings import DATA, EXECUTOR_SETTINGS
    from .manager_commands import send_manager_command

    logger.debug("Collecting files for data object with id {}".format(DATA["id"]))
    reply = await send_manager_command(ExecutorProtocol.GET_REFERENCED_FILES)
    refs = reply[ExecutorProtocol.REFERENCED_FILES]
    base_dir = EXECUTOR_SETTINGS["DATA_DIR"]
    collected, _ = collect_and_purge(base_dir, refs)

    base_dir = Path(base_dir)
    collected_objects = [
        get_transfer_object(base_dir / object_, base_dir) for object_ in collected
    ]

    await send_manager_command(
        ExecutorProtocol.REFERENCED_FILES,
        extra_fields={ExecutorProtocol.REFERENCED_FILES: collected_objects},
    )



"""Storage application views."""

from datetime import datetime
from distutils.util import strtobool
from pathlib import Path
from typing import Union

import pytz

from django.core.exceptions import PermissionDenied
from django.http import HttpRequest, HttpResponse, JsonResponse
from django.shortcuts import redirect
from django.utils import timezone
from django.views import View

from resolwe.flow.models import Data
from resolwe.permissions.shortcuts import get_objects_for_user
from resolwe.storage.models import FileStorage, ReferencedPath


class DataBrowseView(View):
    """Emulate nginx JSON directory view of the data directory.

    Use information in database to construct JSON response that emulates nginx
    autoindex module with autoindex_format set to json. See
    http://nginx.org/en/docs/http/ngx_http_autoindex_module.html
    for details.
    """

    def _get_mtime(self, entry: datetime):
        """Convert datetime entry to nginx autoindex compatible string.

        The timezone returned by autoindex module was set to GMT so we convert
        from our timetone to GMT.
        """
        gmt = timezone.localtime(entry, pytz.timezone("GMT"))
        return gmt.strftime("%a, %d %b %Y %H:%M:%S %Z")

    def _path_to_dict(self, referenced_path: ReferencedPath, base_path: Path):
        """Convert ReferencedPath to dictionary emulating nginx response."""
        data = {
            "name": Path(referenced_path.path).relative_to(base_path).as_posix(),
            "type": "file",
            "mtime": self._get_mtime(referenced_path.file_storage.created),
            "size": referenced_path.size,
        }
        if referenced_path.path.endswith("/"):
            data.pop("size")
            data["type"] = "directory"
        return data

    def get(
        self, request: HttpRequest, *args: Union[str, int], **kwargs: Union[str, int]
    ) -> HttpResponse:
        """Giver URL either redirect to resource or serve a list of resources.

        If URL points to a file then redirect to a default storage location.
        If URL points to a directory then send a JSON encoded list of entires
        in this directory emulating nginx autoindex module.

        :param request: the given request. When the user requests URL
            http://platform-centos.genialis.local/data/1/..., then the value of
            parameter should be 1/... .
        """

        data_id = kwargs.get("data_id")
        relative_path = Path(kwargs.get("uri", ""))

        # Check permissions and load requested data object.
        data_qset = get_objects_for_user(
            self.request.user, "view_data", Data.objects.filter(pk=data_id)
        )
        # Send response with status 403 when requested data object does not exists.
        if not data_qset.exists():
            raise PermissionDenied()

        file_storage: FileStorage = data_qset.get().location
        is_dir = (
            relative_path == Path()
            or file_storage.files.filter(path=relative_path.as_posix() + "/").exists()
        )
        is_file = file_storage.files.filter(path=relative_path.as_posix()).exists()
        if is_dir:
            # Empty path evaluates to "."
            regex_path = (
                relative_path.as_posix() + "/" if relative_path != Path() else ""
            )
            # Show only entries in this directory.
            regex = "^{}[^/]+/?$".format(regex_path)
            data = [
                self._path_to_dict(path, regex_path)
                for path in file_storage.files.filter(path__regex=regex)
            ]
            return JsonResponse(data, safe=False)
        elif is_file:
            # Redirect to the resource.
            force_download = strtobool(request.GET.get("force_download", "false"))
            redirect_url = (
                file_storage.default_storage_location.connector.presigned_url(
                    file_storage.subpath / relative_path, force_download=force_download
                )
            )

            return redirect(redirect_url)
        else:
            raise PermissionDenied()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2018-09-12 08:21
from __future__ import unicode_literals

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0019_relation_type_cicharfield"),
    ]

    operations = [
        migrations.AlterField(
            model_name="process",
            name="category",
            field=models.CharField(
                default="Other:",
                max_length=200,
                validators=[
                    django.core.validators.RegexValidator(
                        code="invalid_category",
                        message="Category may be alphanumerics separated by colon",
                        regex="^([a-zA-Z0-9]+[:\\-])*[a-zA-Z0-9]+:$",
                    )
                ],
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.14 on 2018-08-06 02:28
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0016_refactor_relations_2"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="relationpartition",
            name="position",
        ),
        migrations.AddField(
            model_name="relationpartition",
            name="position",
            field=models.PositiveSmallIntegerField(
                blank=True, db_index=True, null=True
            ),
        ),
    ]



"""Main standalone execution stub, used when the executor is run.

It should be run as a module with one argument: the relative module name
of the concrete executor class to use. The current working directory
should be where the ``executors`` module directory is, so that it can be
imported with python's ``-m <module>`` interpreter option.

Usage format:

.. code-block:: none

    /path/to/python -m executors .executor_type

Concrete example, run from the directory where ``./executors/`` is:

.. code-block:: none

    /venv/bin/python -m executors .docker

using the python from the ``venv`` virtualenv.

.. note::

    The startup code adds the concrete class name as needed, so that in
    the example above, what's actually instantiated is
    ``.docker.run.FlowExecutor``.
"""

import argparse
import asyncio
import logging
import os
import sys
import traceback
from importlib import import_module
from pathlib import Path

from . import manager_commands
from .global_settings import DATA, DATA_META, EXECUTOR_SETTINGS
from .logger import configure_logging
from .manager_commands import send_manager_command
from .protocol import ExecutorProtocol


def handle_exception(exc_type, exc_value, exc_traceback):
    """Log unhandled exceptions."""
    message = "".join(traceback.format_exception(exc_type, exc_value, exc_traceback))
    logger.error("Unhandled exception in executor: {}".format(message))

    loop.run_until_complete(
        asyncio.gather(
            *logging_future_list,
            manager_commands.send_manager_command(
                ExecutorProtocol.UPDATE,
                extra_fields={
                    ExecutorProtocol.UPDATE_CHANGESET: {
                        "process_error": ["Unhandled exception in executor."],
                        "status": DATA_META["STATUS_ERROR"],
                    }
                },
            ),
            manager_commands.send_manager_command(
                ExecutorProtocol.ABORT, expect_reply=False
            ),
        )
    )


sys.excepthook = handle_exception


def get_stdout_json_file():
    """Get exclusive lock over stdout and json file.

    :raises FileExistsException: if the lock can not be obtained.
    """

    def _create_file(filename):
        """Ensure a new file is created and opened for writing."""
        file_descriptor = os.open(filename, os.O_WRONLY | os.O_CREAT | os.O_EXCL)
        return os.fdopen(file_descriptor, "w")

    logger.debug("Preparing output files for Data with id {}".format(DATA["id"]))
    stdout_file = _create_file(Path(EXECUTOR_SETTINGS["DATA_DIR"]) / "stdout.txt")
    json_file = _create_file(Path(EXECUTOR_SETTINGS["DATA_DIR"]) / "jsonout.txt")
    return stdout_file, json_file


async def run_executor(log_file, json_file):
    """Start the actual execution; instantiate the executor and run."""
    parser = argparse.ArgumentParser(description="Run the specified executor.")
    parser.add_argument(
        "module", help="The module from which to instantiate the concrete executor."
    )
    args = parser.parse_args()

    module_name = "{}.run".format(args.module)
    class_name = "FlowExecutor"

    module = import_module(module_name, __package__)
    executor = getattr(module, class_name)()
    with open(ExecutorFiles.PROCESS_SCRIPT, "rt") as script_file:
        await executor.run(DATA["id"], script_file.read(), log_file, json_file)


if __name__ == "__main__":
    # Initialize logging ASAP.
    logging_future_list = []
    loop = asyncio.get_event_loop()
    configure_logging(logging_future_list)
    loop.run_until_complete(manager_commands.init())
    logger = logging.getLogger(__name__)

    # Import storage modules.
    from .collect import collect_files
    from .protocol import ExecutorFiles
    from .transfer import transfer_data

    async def _sequential():
        """Run some things sequentially but asynchronously."""
        try:
            # Try to obtains exclusive lock over stdout and jsonout files.
            # When lock can not be obtained this task is probably a duplicate
            # spawned by celery so lock error and skip processing.
            log_file, json_file = get_stdout_json_file()
            await transfer_data()
            await run_executor(log_file, json_file)
            await collect_files()
        except FileExistsError:
            logger.error("Stdout or jsonout file already exists, aborting.")
            await send_manager_command(ExecutorProtocol.ABORT, expect_reply=False)

    loop.run_until_complete(_sequential())

    # Wait for any pending logging emits now there's
    # nothing else running anymore
    loop.run_until_complete(asyncio.gather(*logging_future_list))

    # Now that logging is done too, close the connection cleanly.
    loop.run_until_complete(manager_commands.deinit())

    # Any stragglers?

    # NOTE: method bellow is deprecated in Python 3.7 in favor of
    # asyncio.all_tasks and will be removed in Python 3.9. The method
    # asycio.all_tasks has been added in Python 3.7 so a switch can be made
    # when support for Python 3.6 is dropped in Resolwe.
    pending = asyncio.Task.all_tasks()
    loop.run_until_complete(asyncio.gather(*pending))

    loop.close()



"""Constants for the Docker executor."""
DATA_VOLUME = "/data"
DATA_ALL_VOLUME = "/data_all"
UPLOAD_VOLUME = "/upload"
SECRETS_VOLUME = "/secrets"

TMPDIR = ".tmp"

# Container image used if no image is specified.
DEFAULT_CONTAINER_IMAGE = "resolwe/base:ubuntu-18.04"



# pylint: disable=missing-docstring
import os
from pathlib import Path

from asgiref.sync import async_to_sync

from django.test import override_settings

from guardian.shortcuts import assign_perm

from resolwe.flow.managers import manager
from resolwe.flow.managers.utils import disable_auto_calls
from resolwe.flow.models import (
    Collection,
    Data,
    DataDependency,
    DescriptorSchema,
    Process,
)
from resolwe.test import ProcessTestCase, TransactionTestCase

PROCESSES_DIR = os.path.join(os.path.dirname(__file__), "processes")


class TestManager(ProcessTestCase):
    def setUp(self):
        super().setUp()

        self.collection = Collection.objects.create(contributor=self.contributor)

        self._register_schemas(processes_paths=[PROCESSES_DIR])

    def test_create_data(self):
        """Test that manager is run when new object is created."""
        process = Process.objects.filter(slug="test-min").latest()
        data = Data.objects.create(
            name="Test data",
            contributor=self.contributor,
            process=process,
        )

        data.refresh_from_db()
        self.assertEqual(data.status, Data.STATUS_DONE)

    def test_spawned_process(self):
        """Test that manager is run for spawned processes and permissions are copied."""
        DescriptorSchema.objects.create(
            name="Test schema", slug="test-schema", contributor=self.contributor
        )
        spawned_process = Process.objects.filter(slug="test-save-file").latest()
        # Patch the process to create Entity, so its bahaviour can be tested.
        spawned_process.entity_type = "test-schema"
        spawned_process.entity_descriptor_schema = "test-schema"
        spawned_process.save()

        assign_perm("view_collection", self.user, self.collection)
        Data.objects.create(
            name="Test data",
            contributor=self.contributor,
            process=Process.objects.filter(slug="test-spawn-new").latest(),
            collection=self.collection,
        )

        # Created and spawned objects should be done.
        self.assertEqual(Data.objects.filter(status=Data.STATUS_DONE).count(), 2)

        # Check that permissions are inherited.
        child = Data.objects.last()
        self.assertTrue(self.user.has_perm("view_data", child))
        self.assertEqual(child.collection.pk, self.collection.pk)
        self.assertEqual(child.entity.collection.pk, self.collection.pk)

    def test_workflow(self):
        """Test that manager is run for workflows."""
        workflow = Process.objects.filter(slug="test-workflow-1").latest()
        data1 = Data.objects.create(
            name="Test data 1",
            contributor=self.contributor,
            process=workflow,
            input={"param1": "world"},
        )
        data2 = Data.objects.create(
            name="Test data 2",
            contributor=self.contributor,
            process=workflow,
            input={"param1": "foobar"},
        )

        # Created and spawned objects should be done.
        self.assertEqual(Data.objects.filter(status=Data.STATUS_DONE).count(), 6)

        # Check correct dependency type is created.
        self.assertEqual(
            {d.kind for d in data1.children_dependency.all()},
            {DataDependency.KIND_SUBPROCESS},
        )
        self.assertEqual(
            {d.kind for d in data2.children_dependency.all()},
            {DataDependency.KIND_SUBPROCESS},
        )

    def test_communicate(self):
        """Make sure communicate can serialize settings.

        When object of PosixPath is used in settings the call to communicate
        fails due to object not being JSON serializable by default.
        """
        with override_settings(FLOW_EXECUTOR={"DATA_DIR": Path("/some/path")}):
            async_to_sync(manager.communicate)(run_sync=True)
        # Make sure original settings are saved.
        async_to_sync(manager.communicate)(run_sync=True)

    def test_dependencies(self):
        """Test that manager handles dependencies correctly."""
        process_parent = Process.objects.filter(slug="test-dependency-parent").latest()
        process_child = Process.objects.filter(slug="test-dependency-child").latest()
        data_parent = Data.objects.create(
            name="Test parent", contributor=self.contributor, process=process_parent
        )
        data_child1 = Data.objects.create(
            name="Test child",
            contributor=self.contributor,
            process=process_child,
            input={},
        )
        data_child2 = Data.objects.create(
            name="Test child",
            contributor=self.contributor,
            process=process_child,
            input={"parent": data_parent.pk},
        )
        data_child3 = Data.objects.create(
            name="Test child",
            contributor=self.contributor,
            process=process_child,
            input={"parent": None},
        )

        data_parent.refresh_from_db()
        data_child1.refresh_from_db()
        data_child2.refresh_from_db()
        data_child3.refresh_from_db()

        # Check locks are created in manager.
        self.assertFalse(data_parent.access_logs.exists())
        self.assertFalse(data_child1.access_logs.exists())
        self.assertTrue(data_child2.access_logs.exists())
        self.assertFalse(data_child3.access_logs.exists())

        # Check that the data_parent location was locked.
        access_log = data_child2.access_logs.get()
        self.assertEqual(
            access_log.storage_location.file_storage.data.get().id, data_parent.id
        )
        # Check that the log is released.
        self.assertIsNotNone(access_log.started)
        self.assertIsNotNone(access_log.finished)

        # Check status.
        self.assertEqual(data_parent.status, Data.STATUS_DONE)
        self.assertEqual(data_child1.status, Data.STATUS_DONE)
        self.assertEqual(data_child2.status, Data.STATUS_DONE)
        self.assertEqual(data_child3.status, Data.STATUS_DONE)

    def test_process_notifications(self):
        process = Process.objects.filter(slug="test-process-notifications").latest()
        data = Data.objects.create(
            name="Test data",
            contributor=self.contributor,
            process=process,
        )

        data.refresh_from_db()

        self.assertEqual(len(data.process_info), 2)
        self.assertEqual(data.process_info[0], "abc")
        self.assertEqual(data.process_info[1][-5:], "xx...")

        self.assertEqual(len(data.process_warning), 1)
        self.assertEqual(data.process_warning[0][-5:], "yy...")

        self.assertEqual(len(data.process_error), 1)
        self.assertEqual(data.process_error[0][-5:], "zz...")


class TransactionTestManager(TransactionTestCase):
    @disable_auto_calls()
    def test_communicate(self):
        process = Process.objects.create(
            name="Input process",
            contributor=self.contributor,
            type="data:test:",
            input_schema=[
                {
                    "name": "input_data",
                    "type": "data:test:",
                    "required": False,
                },
            ],
        )

        data_1 = Data.objects.create(contributor=self.contributor, process=process)
        data_2 = Data.objects.create(
            contributor=self.contributor,
            process=process,
            input={"input_data": data_1.id},
        )
        Data.objects.create(contributor=self.contributor, process=process)
        Data.objects.create(contributor=self.contributor, process=process)

        self.assertEqual(Data.objects.filter(status=Data.STATUS_RESOLVING).count(), 4)

        # Process only one object.
        async_to_sync(manager.communicate)(data_id=data_1.pk, run_sync=True)

        data_1.refresh_from_db()
        self.assertEqual(data_1.status, Data.STATUS_WAITING)
        self.assertEqual(Data.objects.filter(status=Data.STATUS_RESOLVING).count(), 3)

        data_1.status = Data.STATUS_DONE
        data_1.save()

        # Process object's children.
        async_to_sync(manager.communicate)(data_id=data_1.pk, run_sync=True)

        data_2.refresh_from_db()
        self.assertEqual(data_2.status, Data.STATUS_WAITING)
        self.assertEqual(Data.objects.filter(status=Data.STATUS_RESOLVING).count(), 2)

        # Process all objects.
        async_to_sync(manager.communicate)(run_sync=True)

        self.assertEqual(Data.objects.filter(status=Data.STATUS_RESOLVING).count(), 0)



"""Resolwe Python process support."""
from .descriptor import Persistence, ProcessDescriptor, SchedulingClass, ValidationError
from .fields import (
    BooleanField,
    DataField,
    DateField,
    DateTimeField,
    DirField,
    FileField,
    FileHtmlField,
    FloatField,
    GroupField,
    IntegerField,
    JsonField,
    ListField,
    SecretField,
    StringField,
    TextField,
    UrlField,
)
from .runtime import Cmd, Inputs, Outputs, Process

__all__ = (
    "BooleanField",
    "Cmd",
    "DataField",
    "DateField",
    "DateTimeField",
    "DirField",
    "FileField",
    "FileHtmlField",
    "FloatField",
    "GroupField",
    "Inputs",
    "IntegerField",
    "JsonField",
    "ListField",
    "Outputs",
    "Persistence",
    "Process",
    "ProcessDescriptor",
    "SchedulingClass",
    "SecretField",
    "StringField",
    "ValidationError",
    "TextField",
    "UrlField",
)


# XXX: ImportedFormat is available from resolwe-runtime-utils v2.0.0. Docker
# images with older version of resolwe-runtime-utils should skip the import.
try:
    from resolwe_runtime_utils import ImportedFormat

    __all__ += (ImportedFormat,)
except ImportError:
    pass



""".. Ignore pydocstyle D400.

===============
Signal Handlers
===============

"""
from asgiref.sync import async_to_sync

from django.conf import settings
from django.db import transaction
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver

from resolwe.flow.managers import manager
from resolwe.flow.models import Data, Relation
from resolwe.flow.models.entity import RelationPartition


def commit_signal(data_id):
    """Nudge manager at the end of every Data object save event."""
    if not getattr(settings, "FLOW_MANAGER_DISABLE_AUTO_CALLS", False):
        immediate = getattr(settings, "FLOW_MANAGER_SYNC_AUTO_CALLS", False)
        async_to_sync(manager.communicate)(
            data_id=data_id, save_settings=False, run_sync=immediate
        )


@receiver(post_save, sender=Data)
def manager_post_save_handler(sender, instance, created, **kwargs):
    """Run newly created (spawned) processes."""
    if (
        instance.status == Data.STATUS_DONE
        or instance.status == Data.STATUS_ERROR
        or created
    ):
        # Run manager at the end of the potential transaction. Otherwise
        # tasks are send to workers before transaction ends and therefore
        # workers cannot access objects created inside transaction.
        transaction.on_commit(lambda: commit_signal(instance.id))


# NOTE: m2m_changed signal cannot be used because of a bug:
# https://code.djangoproject.com/ticket/17688
@receiver(post_delete, sender=RelationPartition)
def delete_relation(sender, instance, **kwargs):
    """Delete the Relation object when the last Entity is removed."""

    def process_signal(relation_id):
        """Get the relation and delete it if it has no entities left."""
        try:
            relation = Relation.objects.get(pk=relation_id)
        except Relation.DoesNotExist:
            return

        if relation.entities.count() == 0:
            relation.delete()

    # Wait for partitions to be recreated.
    transaction.on_commit(lambda: process_signal(instance.relation_id))



# pylint: disable=missing-docstring
from guardian.models import GroupObjectPermission, UserObjectPermission
from guardian.shortcuts import assign_perm

from resolwe.flow.models import Collection, DescriptorSchema, Process
from resolwe.permissions.utils import (
    change_perm_ctype,
    copy_permissions,
    get_full_perm,
    get_perm_action,
)
from resolwe.test import TestCase


class UtilsTestCase(TestCase):
    def setUp(self):
        super().setUp()

        self.src_process = Process.objects.create(
            name="Source process", contributor=self.contributor
        )
        self.dst_process = Process.objects.create(
            name="Destination process", contributor=self.contributor
        )

        self.collection = Collection.objects.create(
            name="Test collection", contributor=self.contributor
        )

        self.descriptor_schema = DescriptorSchema.objects.create(
            contributor=self.contributor
        )

    def test_get_perm_action(self):
        self.assertEqual(get_perm_action("view_data"), "view")
        self.assertEqual(get_perm_action("view"), "view")

    def test_get_full_perm(self):
        self.assertEqual(get_full_perm("view", self.collection), "view_collection")
        self.assertEqual(
            get_full_perm("view", self.descriptor_schema), "view_descriptorschema"
        )

    def test_change_perm_ctype(self):
        self.assertEqual(
            change_perm_ctype("view_data", self.collection), "view_collection"
        )

    def test_copy_permissions(self):
        assign_perm("view_process", self.contributor, self.src_process)
        assign_perm("view_process", self.group, self.src_process)

        copy_permissions(self.src_process, self.dst_process)

        self.assertEqual(GroupObjectPermission.objects.count(), 2)
        self.assertEqual(UserObjectPermission.objects.count(), 2)

        self.assertTrue(
            self.contributor.has_perm("flow.view_process", self.dst_process)
        )
        # User inherites permission from group
        self.assertTrue(self.user.has_perm("flow.view_process", self.dst_process))

    def test_copy_different_ctype(self):
        assign_perm("view_collection", self.contributor, self.collection)
        assign_perm("edit_collection", self.contributor, self.collection)

        copy_permissions(self.collection, self.dst_process)

        # Only 'view' is copied as process has no 'add' permission.
        self.assertEqual(UserObjectPermission.objects.count(), 3)



"""Python process runner utility."""
import argparse
import inspect
import json
import os
import sys
from importlib import import_module

from .descriptor import ValidationError
from .runtime import Inputs, Process

if __name__ == "__main__":

    parser = argparse.ArgumentParser(description="Run a Resolwe Python proces")
    parser.add_argument("filename", type=str, help="Python process filename to run")
    parser.add_argument(
        "--slug",
        type=str,
        help="Slug of the process to run (required if multiple processes are defined)",
    )
    parser.add_argument(
        "--name", type=str, help="Name of processing Data object", default=""
    )
    parser.add_argument("--inputs", type=str, help="Path to input JSON file")
    parser.add_argument(
        "--requirements", type=str, help="Path to requirements JSON file"
    )
    parser.add_argument("--relations", type=str, help="Path to relations JSON file")
    args = parser.parse_args()

    # Switch to target directory to import the module.
    try:
        filename = os.path.realpath(args.filename)
        start_dir = os.getcwd()
        os.chdir(os.path.dirname(filename))
        module, _ = os.path.splitext(os.path.basename(filename))
        module = import_module(module, __package__)
        os.chdir(start_dir)
    except (OSError, ImportError):
        print("ERROR: Failed to load Python process from '{}'.".format(args.filename))
        sys.exit(1)

    processes = {}
    for variable in dir(module):
        value = getattr(module, variable)
        if (
            value == Process
            or not inspect.isclass(value)
            or not issubclass(value, Process)
        ):
            continue

        processes[value._meta.metadata.slug] = value

    if args.slug:
        try:
            process = processes[args.slug]
        except KeyError:
            print("Found the following processes in module '{}':".format(args.module))
            print("")
            for slug, process in processes.items():
                print("  {} ({})".format(slug, process._meta.metadata.name))
            print("")

            print("ERROR: Unable to find process '{}'.".format(args.slug))
            sys.exit(1)
    elif len(processes) == 1:
        process = next(iter(processes.values()))
    else:
        print("Found the following processes in module '{}':".format(args.module))
        print("")
        for slug, process in processes.items():
            print("  {} ({})".format(slug, process._meta.metadata.name))
        print("")

        print("ERROR: Unable to determine which process to run. Pass --slug option.")
        sys.exit(1)

    # Prepare process inputs.
    try:
        inputs = Inputs(process._meta.inputs)
        if args.inputs:
            with open(args.inputs) as inputs_file:
                data = json.load(inputs_file)

            for key, value in data.items():
                setattr(inputs, key, value)
    except ValidationError as error:
        print("ERROR: Input field validation failed: {}".format(error.args[0]))
        sys.exit(1)

    # Prepare process requirements.
    requirements = None
    try:
        if args.requirements:
            with open(args.requirements) as requirements_file:
                requirements = json.load(requirements_file)
    except Exception as error:
        print("ERROR: Requirements failed to load: {}".format(str(error)))
        sys.exit(1)

    # Prepare relations.
    relations = None
    try:
        if args.relations:
            with open(args.relations) as relations_file:
                relations = json.load(relations_file)
    except Exception as error:
        print("ERROR: Relations failed to load: {}".format(str(error)))
        sys.exit(1)

    # TODO: Configure logging.

    # Start the process.
    instance = process()
    try:
        if requirements is not None:
            instance._meta.metadata.requirements = requirements
        instance._meta.relations = relations
        instance._meta.name = args.name

        instance.start(inputs)
    except ValidationError as error:
        print("ERROR: Output field validation failed: {}".format(error.args[0]))
        sys.exit(1)



import os

# -- General configuration ------------------------------------------------

# The extension modules to enable.
extensions = ["resolwe.flow.utils.docs.autoprocess"]

# Parent directory of all process definitions:
autoprocess_process_dir = os.path.normpath(
    os.path.abspath(os.path.dirname(__file__)) + "/processes"
)

# Base of the url to process source code:
autoprocess_source_base_url = (
    "https://github.com/genialis/resolwe-bio/blob/master/resolwe_bio/processes/"
)
